# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement
import asyncio
import re
import time

# rediscluster imports
from coredis import StrictRedisCluster, StrictRedis
from coredis.pool import ClusterConnectionPool
from coredis.exceptions import (
    RedisClusterException,
    MovedError,
    AskError,
    ClusterDownError,
)
from coredis.utils import b
from coredis.nodemanager import NodeManager
from tests.cluster.conftest import _get_client, skip_if_not_password_protected_nodes

# 3rd party imports
from mock import patch, Mock, MagicMock
import pytest


class DummyConnectionPool(ClusterConnectionPool):
    pass


class DummyConnection:
    pass


# def test_representation(r):
#     assert re.search('^StrictRedisCluster<[0-9\.\:\,].+>$', str(r))
#
#
# def test_blocked_strict_redis_args():
#     """
#     Some arguments should explicitly be blocked because they will not work in a cluster setup
#     """
#     params = {'startup_nodes': [{'host': '127.0.0.1', 'port': 7000}]}
#     c = StrictRedisCluster(**params)
#     assert c.connection_pool.connection_kwargs["stream_timeout"] == ClusterConnectionPool.RedisClusterDefaultTimeout
#
#     with pytest.raises(RedisClusterException) as ex:
#         _get_client(db=1)
#     assert str(ex.value).startswith("Argument 'db' is not possible to use in cluster mode")
#
#
# @skip_if_not_password_protected_nodes()
# def test_password_procted_nodes():
#     """
#     Test that it is possible to connect to password protected nodes
#     """
#     startup_nodes = [{"host": "127.0.0.1", "port": "7000"}]
#     password_protected_startup_nodes = [{"host": "127.0.0.1", "port": "7100"}]
#     with pytest.raises(RedisClusterException) as ex:
#         _get_client(startup_nodes=password_protected_startup_nodes)
#     assert str(ex.value).startswith("ERROR sending 'cluster slots' command to redis server:")
#     _get_client(startup_nodes=password_protected_startup_nodes, password='password_is_protected')
#
#     with pytest.raises(RedisClusterException) as ex:
#         _get_client(startup_nodes=startup_nodes, password='password_is_protected')
#     assert str(ex.value).startswith("ERROR sending 'cluster slots' command to redis server:")
#     _get_client(startup_nodes=startup_nodes)
#
#
# def test_host_port_startup_node():
#     """
#     Test that it is possible to use host & port arguments as startup node args
#     """
#     h = "192.168.0.1"
#     p = 7000
#     c = StrictRedisCluster(host=h, port=p)
#     assert {"host": h, "port": p} in c.connection_pool.nodes.startup_nodes
#
#
# def test_empty_startup_nodes():
#     """
#     Test that exception is raised when empty providing empty startup_nodes
#     """
#     with pytest.raises(RedisClusterException) as ex:
#         _get_client(init_slot_cache=False, startup_nodes=[])
#
#     assert str(ex.value).startswith("No startup nodes provided"), str(ex.value)
#
#
# def test_readonly_instance(ro):
#     """
#     Test that readonly_mode=True instance has ClusterReadOnlyConnectionPool
#     """
#     assert ro.connection_pool.readonly is True
#
#
# def test_custom_connectionpool():
#     """
#     Test that a custom connection pool will be used by StrictRedisCluster
#     """
#     h = "192.168.0.1"
#     p = 7001
#     pool = DummyConnectionPool(host=h, port=p, connection_class=DummyConnection,
#                                startup_nodes=[{'host': h, 'port': p}])
#     c = StrictRedisCluster(connection_pool=pool)
#     assert c.connection_pool is pool
#     assert c.connection_pool.connection_class == DummyConnection
#     assert {"host": h, "port": p} in c.connection_pool.nodes.startup_nodes
#
#
# @patch('coredis.StrictRedis', new=MagicMock())
# def test_skip_full_coverage_check():
#     """
#     Test if the cluster_require_full_coverage NodeManager method was not called with the flag activated
#     """
#     c = StrictRedisCluster("192.168.0.1", 7001, skip_full_coverage_check=True)
#     c.connection_pool.nodes.cluster_require_full_coverage = MagicMock()
#     c.connection_pool.nodes.initialize()
#     assert not c.connection_pool.nodes.cluster_require_full_coverage.called
#
#
# @pytest.mark.asyncio
# async def test_blocked_commands(r):
#     """
#     These commands should be blocked and raise RedisClusterException
#     """
#     blocked_commands = [
#         "CLIENT SETNAME", "SENTINEL GET-MASTER-ADDR-BY-NAME", 'SENTINEL MASTER', 'SENTINEL MASTERS',
#         'SENTINEL MONITOR', 'SENTINEL REMOVE', 'SENTINEL SENTINELS', 'SENTINEL SET',
#         'SENTINEL SLAVES', 'SHUTDOWN', 'SLAVEOF', 'SCRIPT KILL', 'MOVE', 'BITOP',
#     ]
#
#     for command in blocked_commands:
#         try:
#             await r.execute_command(command)
#         except RedisClusterException:
#             pass
#         else:
#             raise AssertionError("'RedisClusterException' not raised for method : {0}".format(command))
#
# @pytest.mark.asyncio
# async def test_blocked_transaction(r):
#     """
#     Method transaction is blocked/NYI and should raise exception on use
#     """
#     with pytest.raises(RedisClusterException) as ex:
#         await r.transaction(None)
#     assert str(ex.value).startswith("method StrictRedisCluster.transaction() is not implemented"), str(ex.value)
#
#
# @pytest.mark.asyncio
# async def test_cluster_of_one_instance():
#     """
#     Test a cluster that starts with only one redis server and ends up with
#     one server.
#
#     There is another redis server joining the cluster, hold slot 0, and
#     eventually quit the cluster. The StrictRedisCluster instance may get confused
#     when slots mapping and nodes change during the test.
#     """
#     with patch.object(StrictRedisCluster, 'parse_response') as parse_response_mock:
#         with patch.object(NodeManager, 'initialize', autospec=True) as init_mock:
#             async def side_effect(self, *args, **kwargs):
#                 async def ok_call(self, *args, **kwargs):
#                     assert self.port == 7007
#                     return "OK"
#                 parse_response_mock.side_effect = ok_call
#
#                 raise ClusterDownError('CLUSTERDOWN The cluster is down. Use CLUSTER INFO for more information')
#
#             async def side_effect_rebuild_slots_cache(self):
#                 # make new node cache that points to 7007 instead of 7006
#                 self.nodes = [{'host': '127.0.0.1', 'server_type': 'master', 'port': 7006, 'name': '127.0.0.1:7006'}]
#                 self.slots = {}
#
#                 for i in range(0, 16383):
#                     self.slots[i] = [{
#                         'host': '127.0.0.1',
#                         'server_type': 'master',
#                         'port': 7006,
#                         'name': '127.0.0.1:7006',
#                     }]
#
#                 # Second call should map all to 7007
#                 async def map_7007(self):
#                     self.nodes = [{'host': '127.0.0.1', 'server_type': 'master', 'port': 7007, 'name': '127.0.0.1:7007'}]
#                     self.slots = {}
#
#                     for i in range(0, 16383):
#                         self.slots[i] = [{
#                             'host': '127.0.0.1',
#                             'server_type': 'master',
#                             'port': 7007,
#                             'name': '127.0.0.1:7007',
#                         }]
#
#                 # First call should map all to 7006
#                 init_mock.side_effect = map_7007
#
#             parse_response_mock.side_effect = side_effect
#             init_mock.side_effect = side_effect_rebuild_slots_cache
#
#             rc = StrictRedisCluster(host='127.0.0.1', port=7006)
#             await rc.set('bar', 'foo')
#
#
# @pytest.mark.asyncio
# async def test_execute_command_errors(r):
#     """
#     If no command is given to `_determine_nodes` then exception
#     should be raised.
#
#     Test that if no key is provided then exception should be raised.
#     """
#     with pytest.raises(RedisClusterException) as ex:
#         await r.execute_command()
#     assert str(ex.value).startswith("Unable to determine command to use")
#
#     with pytest.raises(RedisClusterException) as ex:
#         await r.execute_command("GET")
#     assert str(ex.value).startswith("No way to dispatch this command to Redis Cluster. Missing key.")
#
#
# @pytest.mark.asyncio
# async def test_refresh_table_asap(monkeypatch):
#     """
#     If this variable is set externally, initialize() should be called.
#     """
#     async def return_none(*args, **kwargs):
#         return None
#     monkeypatch.setattr(ClusterConnectionPool, 'initialize', return_none)
#     monkeypatch.setattr(StrictRedisCluster, 'parse_response', return_none)
#     r = StrictRedisCluster(host="127.0.0.1", port=7000)
#     r.connection_pool.nodes.slots[12182] = [{
#         "host": "127.0.0.1",
#         "port": 7002,
#         "name": "127.0.0.1:7002",
#         "server_type": "master",
#     }]
#     r.refresh_table_asap = True
#
#     await r.set("foo", "bar")
#     assert r.refresh_table_asap is False
#     await r.flushdb()
#
#
# def find_node_ip_based_on_port(cluster_client, port):
#     for node_name, node_data in cluster_client.connection_pool.nodes.nodes.items():
#         if node_name.endswith(port):
#             return node_data['host']
#
# @pytest.mark.asyncio
# async def test_ask_redirection():
#     """
#     Test that the server handles ASK response.
#
#     At first call it should return a ASK ResponseError that will point
#     the client to the next server it should talk to.
#
#     Important thing to verify is that it tries to talk to the second node.
#     """
#     r = StrictRedisCluster(host="127.0.0.1", port=7000)
#     await r.connection_pool.initialize()
#     r.connection_pool.nodes.nodes['127.0.0.1:7001'] = {
#         'host': u'127.0.0.1',
#         'server_type': 'master',
#         'port': 7001,
#         'name': '127.0.0.1:7001'
#     }
#     with patch.object(StrictRedisCluster,
#                       'parse_response') as parse_response:
#
#         host_ip = find_node_ip_based_on_port(r, '7001')
#
#         async def ask_redirect_effect(connection, *args, **options):
#             async def ok_response(connection, *args, **options):
#                 assert connection.host == host_ip
#                 assert connection.port == 7001
#
#                 return "MOCK_OK"
#             parse_response.side_effect = ok_response
#             raise AskError("1337 {0}:7001".format(host_ip))
#
#         parse_response.side_effect = ask_redirect_effect
#         assert await r.set("foo", "bar") == "MOCK_OK"


@pytest.mark.asyncio
async def test_pipeline_ask_redirection():
    """
    Test that the server handles ASK response when used in pipeline.

    At first call it should return a ASK ResponseError that will point
    the client to the next server it should talk to.

    Important thing to verify is that it tries to talk to the second node.
    """
    r = StrictRedisCluster(host="127.0.0.1", port=7000)
    with patch.object(StrictRedisCluster, "parse_response") as parse_response:

        async def response(connection, *args, **options):
            async def response(connection, *args, **options):
                async def response(connection, *args, **options):
                    assert connection.host == "127.0.0.1"
                    assert connection.port == 7001
                    return "MOCK_OK"

                parse_response.side_effect = response
                raise AskError("12182 127.0.0.1:7001")

            parse_response.side_effect = response
            raise AskError("12182 127.0.0.1:7001")

        parse_response.side_effect = response

        p = await r.pipeline()
        await p.connection_pool.initialize()
        p.connection_pool.nodes.nodes["127.0.0.1:7001"] = {
            "host": u"127.0.0.1",
            "server_type": "master",
            "port": 7001,
            "name": "127.0.0.1:7001",
        }
        await p.set("foo", "bar")
        assert await p.execute() == ["MOCK_OK"]


@pytest.mark.asyncio
async def test_moved_redirection():
    """
    Test that the client handles MOVED response.

    At first call it should return a MOVED ResponseError that will point
    the client to the next server it should talk to.

    Important thing to verify is that it tries to talk to the second node.
    """
    r0 = StrictRedisCluster(host="127.0.0.1", port=7000)
    r2 = StrictRedisCluster(host="127.0.0.1", port=7002)
    await r0.flushdb()
    await r2.flushdb()
    assert await r0.set("foo", "bar")
    assert await r2.get("foo") == b"bar"


@pytest.mark.asyncio
async def test_moved_redirection_pipeline(monkeypatch):
    """
    Test that the server handles MOVED response when used in pipeline.

    At first call it should return a MOVED ResponseError that will point
    the client to the next server it should talk to.

    Important thing to verify is that it tries to talk to the second node.
    """
    r0 = StrictRedisCluster(host="127.0.0.1", port=7000)
    r2 = StrictRedisCluster(host="127.0.0.1", port=7002)
    await r0.flushdb()
    await r2.flushdb()
    p = await r0.pipeline()
    await p.set("foo", "bar")
    assert await p.execute() == [True]
    assert await r2.get("foo") == b"bar"


async def assert_moved_redirection_on_slave(sr, connection_pool_cls, cluster_obj):
    """
    """
    # we assume this key is set on 127.0.0.1:7000(7003)
    await sr.set("foo16706", "foo")
    await asyncio.sleep(1)
    with patch.object(connection_pool_cls, "get_node_by_slot") as return_slave_mock:
        return_slave_mock.return_value = {
            "name": "127.0.0.1:7004",
            "host": "127.0.0.1",
            "port": 7004,
            "server_type": "slave",
        }

        master_value = {
            "host": "127.0.0.1",
            "name": "127.0.0.1:7000",
            "port": 7000,
            "server_type": "master",
        }
        with patch.object(
            ClusterConnectionPool, "get_master_node_by_slot"
        ) as return_master_mock:
            return_master_mock.return_value = master_value
            assert await cluster_obj.get("foo16706") == b("foo")
            assert return_slave_mock.call_count == 1


@pytest.mark.asyncio
async def test_moved_redirection_on_slave_with_default_client(sr):
    """
    Test that the client is redirected normally with default
    (readonly_mode=False) client even when we connect always to slave.
    """
    await assert_moved_redirection_on_slave(
        sr,
        ClusterConnectionPool,
        StrictRedisCluster(host="127.0.0.1", port=7000, reinitialize_steps=1),
    )


@pytest.mark.asyncio
async def test_moved_redirection_on_slave_with_readonly_mode_client(sr):
    """
    Ditto with READONLY mode.
    """
    await assert_moved_redirection_on_slave(
        sr,
        ClusterConnectionPool,
        StrictRedisCluster(
            host="127.0.0.1", port=7000, readonly=True, reinitialize_steps=1
        ),
    )


@pytest.mark.asyncio
async def test_access_correct_slave_with_readonly_mode_client(sr):
    """
    Test that the client can get value normally with readonly mode
    when we connect to correct slave.
    """

    # we assume this key is set on 127.0.0.1:7000(7003)
    await sr.set("foo16706", "foo")
    await asyncio.sleep(1)

    with patch.object(ClusterConnectionPool, "get_node_by_slot") as return_slave_mock:
        return_slave_mock.return_value = {
            "name": "127.0.0.1:7004",
            "host": "127.0.0.1",
            "port": 7004,
            "server_type": "slave",
        }

        master_value = {
            "host": "127.0.0.1",
            "name": "127.0.0.1:7000",
            "port": 7000,
            "server_type": "master",
        }
        with patch.object(
            ClusterConnectionPool, "get_master_node_by_slot", return_value=master_value
        ) as return_master_mock:
            readonly_client = StrictRedisCluster(
                host="127.0.0.1", port=7000, readonly=True
            )
            assert b("foo") == await readonly_client.get("foo16706")
            readonly_client = StrictRedisCluster.from_url(
                url="redis://127.0.0.1:7000/0", readonly=True
            )
            assert b("foo") == await readonly_client.get("foo16706")



# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement
import re

# rediscluster imports
from coredis import StrictRedisCluster, ClusterConnectionPool
from coredis.utils import b
from coredis.exceptions import (
    RedisClusterException,
    WatchError,
    ResponseError,
    ConnectionError,
)
from tests.cluster.conftest import _get_client

# 3rd party imports
import pytest
from mock import patch


class TestPipeline:
    """
    """

    @pytest.mark.asyncio()
    async def test_pipeline(self, r):
        await r.flushdb()
        async with await r.pipeline() as pipe:
            await (
                await (await (await pipe.set("a", "a1")).get("a")).zadd("z", z1=1)
            ).zadd("z", z2=4)
            await (await pipe.zincrby("z", "z1")).zrange("z", 0, 5, withscores=True)
            assert await pipe.execute() == [
                True,
                b("a1"),
                True,
                True,
                2.0,
                [(b("z1"), 2.0), (b("z2"), 4)],
            ]

    @pytest.mark.asyncio()
    async def test_pipeline_length(self, r):
        await r.flushdb()
        async with await r.pipeline() as pipe:
            # Initially empty.
            assert len(pipe) == 0
            assert not pipe

            # Fill 'er up!
            await (await (await pipe.set("a", "a1")).set("b", "b1")).set("c", "c1")
            assert len(pipe) == 3
            assert pipe

            # Execute calls reset(), so empty once again.
            await pipe.execute()
            assert len(pipe) == 0
            assert not pipe

    @pytest.mark.asyncio()
    async def test_pipeline_no_transaction(self, r):
        await r.flushdb()
        async with await r.pipeline(transaction=False) as pipe:
            await (await (await pipe.set("a", "a1")).set("b", "b1")).set("c", "c1")
            assert await pipe.execute() == [True, True, True]
            assert await r.get("a") == b("a1")
            assert await r.get("b") == b("b1")
            assert await r.get("c") == b("c1")

    @pytest.mark.asyncio()
    async def test_pipeline_eval(self, r):
        await r.flushdb()
        async with await r.pipeline(transaction=False) as pipe:
            await pipe.eval(
                "return {KEYS[1],KEYS[2],ARGV[1],ARGV[2]}",
                2,
                "A{foo}",
                "B{foo}",
                "first",
                "second",
            )
            res = (await pipe.execute())[0]
            assert res[0] == b("A{foo}")
            assert res[1] == b("B{foo}")
            assert res[2] == b("first")
            assert res[3] == b("second")

    @pytest.mark.asyncio()
    @pytest.mark.xfail(reason="unsupported command: watch")
    async def test_pipeline_no_transaction_watch(self, r):
        await r.flushdb()
        await r.set("a", 0)

        async with await r.pipeline(transaction=False) as pipe:
            await pipe.watch("a")
            a = await pipe.get("a")

            await pipe.multi()
            await pipe.set("a", int(a) + 1)
            assert await pipe.execute() == [True]

    @pytest.mark.asyncio()
    @pytest.mark.xfail(reason="unsupported command: watch")
    async def test_pipeline_no_transaction_watch_failure(self, r):
        await r.flushdb()
        await r.set("a", 0)

        async with await r.pipeline(transaction=False) as pipe:
            await pipe.watch("a")
            a = await pipe.get("a")

            await r.set("a", "bad")

            await pipe.multi()
            await pipe.set("a", int(a) + 1)

            with pytest.raises(WatchError):
                await pipe.execute()

            assert await r.get("a") == b("bad")

    @pytest.mark.asyncio()
    async def test_exec_error_in_response(self, r):
        """
        an invalid pipeline command at exec time adds the exception instance
        to the list of returned values
        """
        await r.flushdb()
        await r.set("c", "a")
        async with await r.pipeline() as pipe:
            await (
                await (await (await pipe.set("a", 1)).set("b", 2)).lpush("c", 3)
            ).set("d", 4)
            result = await pipe.execute(raise_on_error=False)

            assert result[0]
            assert await r.get("a") == b("1")
            assert result[1]
            assert await r.get("b") == b("2")

            # we can't lpush to a key that's a string value, so this should
            # be a ResponseError exception
            assert isinstance(result[2], ResponseError)
            assert await r.get("c") == b("a")

            # since this isn't a transaction, the other commands after the
            # error are still executed
            assert result[3]
            assert await r.get("d") == b("4")

            # make sure the pipe was restored to a working state
            await pipe.set("z", "zzz")
            assert await pipe.execute() == [True]
            assert await r.get("z") == b("zzz")

    @pytest.mark.asyncio()
    async def test_exec_error_raised(self, r):
        await r.flushdb()
        await r.set("c", "a")
        async with await r.pipeline() as pipe:
            await pipe.set("a", 1)
            await pipe.set("b", 2)
            await pipe.lpush("c", 3)
            await pipe.set("d", 4)
            with pytest.raises(ResponseError) as ex:
                await pipe.execute()
            assert str(ex.value).startswith(
                "Command # 3 (LPUSH c 3) of " "pipeline caused error: "
            )

            # make sure the pipe was restored to a working state
            await pipe.set("z", "zzz")
            assert await pipe.execute() == [True]
            assert await r.get("z") == b("zzz")

    @pytest.mark.asyncio()
    async def test_parse_error_raised(self, r):
        async with await r.pipeline() as pipe:
            # the zrem is invalid because we don't pass any keys to it
            await (await (await pipe.set("a", 1)).zrem("b")).set("b", 2)
            with pytest.raises(ResponseError) as ex:
                await pipe.execute()

            assert str(ex.value).startswith(
                "Command # 2 (ZREM b) of pipeline caused error: "
            )

            # make sure the pipe was restored to a working state
            assert await (await pipe.set("z", "zzz")).execute() == [True]
            assert await r.get("z") == b("zzz")

    @pytest.mark.asyncio()
    @pytest.mark.xfail(reason="unsupported command: watch")
    async def test_watch_succeed(self, r):
        await r.flushdb()
        await r.set("a", 1)
        await r.set("b", 2)

        async with await r.pipeline() as pipe:
            await pipe.watch("a", "b")
            assert pipe.watching
            a_value = await pipe.get("a")
            b_value = await pipe.get("b")
            assert a_value == b("1")
            assert b_value == b("2")
            await pipe.multi()

            await pipe.set("c", 3)
            assert await pipe.execute() == [True]
            assert not pipe.watching

    @pytest.mark.asyncio()
    @pytest.mark.xfail(reason="unsupported command: watch")
    async def test_watch_failure(self, r):
        await r.flushdb()
        await r.set("a", 1)
        await r.set("b", 2)

        async with await r.pipeline() as pipe:
            await pipe.watch("a", "b")
            r["b"] = 3
            await pipe.multi()
            await pipe.get("a")
            with pytest.raises(WatchError):
                await pipe.execute()

            assert not pipe.watching

    @pytest.mark.asyncio()
    @pytest.mark.xfail(reason="unsupported command: watch")
    async def test_unwatch(self, r):
        await r.flushdb()
        await r.set("a", 1)
        await r.set("b", 2)

        async with await r.pipeline() as pipe:
            await pipe.watch("a", "b")
            r["b"] = 3
            await pipe.unwatch()
            assert not pipe.watching
            await pipe.get("a")
            assert await pipe.execute() == [b("1")]

    @pytest.mark.asyncio()
    @pytest.mark.xfail(reason="unsupported command: watch")
    async def test_transaction_callable(self, r):
        await r.flushdb()
        await r.set("a", 1)
        await r.set("b", 2)
        has_run = []

        async def my_transaction(pipe):
            a_value = pipe.get("a")
            assert a_value in (b("1"), b("2"))
            b_value = pipe.get("b")
            assert b_value == b("2")

            # silly run-once code... incr's "a" so WatchError should be raised
            # forcing this all to run again. this should incr "a" once to "2"
            if not has_run:
                await r.incr("a")
                has_run.append("it has")

            await pipe.multi()
            await pipe.set("c", int(a_value) + int(b_value))

        result = await r.transaction(my_transaction, "a", "b")
        assert result == [True]
        assert await r.get("c") == b("4")


#
#     def test_exec_error_in_no_transaction_pipeline(self, r):
#         r['a'] = 1
#         with r.pipeline(transaction=False) as pipe:
#             pipe.llen('a')
#             pipe.expire('a', 100)
#
#             with pytest.raises(ResponseError) as ex:
#                 pipe.execute()
#
#             assert ex.value.startswith('Command # 1 (LLEN a) of '
#                                                 'pipeline caused error: ')
#
#         assert r['a'] == b('1')
#
#     def test_exec_error_in_no_transaction_pipeline_unicode_command(self, r):
#         key = chr(3456) + 'abcd' + chr(3421)
#         r[key] = 1
#         with r.pipeline(transaction=False) as pipe:
#             pipe.llen(key)
#             pipe.expire(key, 100)
#
#             with pytest.raises(ResponseError) as ex:
#                 pipe.execute()
#
#             expected = str('Command # 1 (LLEN {0}) of pipeline caused error: ').format(key)
#             assert ex.value.startswith(expected)
#
#         assert r[key] == b('1')
#
#     def test_blocked_methods(self, r):
#         """
#         Currently some method calls on a Cluster pipeline
#         is blocked when using in cluster mode.
#         They maybe implemented in the future.
#         """
#         pipe = r.pipeline(transaction=False)
#         with pytest.raises(RedisClusterException):
#             pipe.multi()
#
#         with pytest.raises(RedisClusterException):
#             pipe.immediate_execute_command()
#
#         with pytest.raises(RedisClusterException):
#             pipe._execute_transaction(None, None, None)
#
#         with pytest.raises(RedisClusterException):
#             pipe.load_scripts()
#
#         with pytest.raises(RedisClusterException):
#             pipe.watch()
#
#         with pytest.raises(RedisClusterException):
#             pipe.unwatch()
#
#         with pytest.raises(RedisClusterException):
#             pipe.script_load_for_pipeline(None)
#
#         with pytest.raises(RedisClusterException):
#             pipe.transaction(None)
#
#     def test_blocked_arguments(self, r):
#         """
#         Currently some arguments is blocked when using in cluster mode.
#         They maybe implemented in the future.
#         """
#         with pytest.raises(RedisClusterException) as ex:
#             r.pipeline(transaction=True)
#
#         assert ex.value.startswith("transaction is deprecated in cluster mode"), True
#
#         with pytest.raises(RedisClusterException) as ex:
#             r.pipeline(shard_hint=True)
#
#         assert ex.value.startswith("shard_hint is deprecated in cluster mode"), True
#
#     def test_redis_cluster_pipeline(self):
#         """
#         Test that we can use a pipeline with the RedisCluster class
#         """
#         r = _get_client(cls=None)
#         with r.pipeline(transaction=False) as pipe:
#             pipe.get("foobar")
#
#     def test_mget_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.mget(['a'])
#
#     def test_mset_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.mset({'a': 1, 'b': 2})
#
#     def test_rename_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.rename('a', 'b')
#
#     def test_renamenx_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.renamenx('a', 'b')
#
#     def test_delete_single(self, r):
#         r['a'] = 1
#         with r.pipeline(transaction=False) as pipe:
#             pipe.delete('a')
#             assert pipe.execute(), True
#
#     def test_multi_delete_unsupported(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             r['a'] = 1
#             r['b'] = 2
#             with pytest.raises(RedisClusterException):
#                 pipe.delete('a', 'b')
#
#     def test_brpoplpush_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.brpoplpush()
#
#     def test_rpoplpush_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.rpoplpush()
#
#     def test_sort_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.sort()
#
#     def test_sdiff_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.sdiff()
#
#     def test_sdiffstore_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.sdiffstore()
#
#     def test_sinter_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.sinter()
#
#     def test_sinterstore_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.sinterstore()
#
#     def test_smove_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.smove()
#
#     def test_sunion_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.sunion()
#
#     def test_sunionstore_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.sunionstore()
#
#     def test_spfmerge_disabled(self, r):
#         with r.pipeline(transaction=False) as pipe:
#             with pytest.raises(RedisClusterException):
#                 pipe.pfmerge()
#
#     def test_multi_key_operation_with_shared_shards(self, r):
#         pipe = r.pipeline(transaction=False)
#         pipe.set('a{foo}', 1)
#         pipe.set('b{foo}', 2)
#         pipe.set('c{foo}', 3)
#         pipe.set('bar', 4)
#         pipe.set('bazz', 5)
#         pipe.get('a{foo}')
#         pipe.get('b{foo}')
#         pipe.get('c{foo}')
#         pipe.get('bar')
#         pipe.get('bazz')
#         res = pipe.execute()
#         assert res == [True, True, True, True, True, b'1', b'2', b'3', b'4', b'5']
#
#     @pytest.mark.xfail(reson="perform_execute_pipeline is not used any longer")
#     def test_connection_error(self, r):
#         test = self
#         test._calls = []
#
#         def perform_execute_pipeline(pipe):
#             if not test._calls:
#                 e = ConnectionError('test')
#                 test._calls.append({'exception': e})
#                 return [e]
#             result = pipe.execute(raise_on_error=False)
#             test._calls.append({'result': result})
#             return result
#
#         pipe = r.pipeline(transaction=False)
#         orig_perform_execute_pipeline = pipe.perform_execute_pipeline
#         pipe.perform_execute_pipeline = perform_execute_pipeline
#
#         try:
#             pipe.set('foo', 1)
#             res = pipe.execute()
#             assert res, [True]
#             assert isinstance(test._calls[0]['exception'], ConnectionError)
#             if len(test._calls) == 2:
#                 assert test._calls[1] == {'result': [True]}
#             else:
#                 assert isinstance(test._calls[1]['result'][0], ResponseError)
#                 assert test._calls[2] == {'result': [True]}
#         finally:
#             pipe.perform_execute_pipeline = orig_perform_execute_pipeline
#             del test._calls
#
#     @pytest.mark.xfail(reson="perform_execute_pipeline is not used any longer")
#     def test_asking_error(self, r):
#         test = self
#         test._calls = []
#
#         def perform_execute_pipeline(pipe):
#             if not test._calls:
#
#                 e = ResponseError("ASK {0} 127.0.0.1:7003".format(r.keyslot('foo')))
#                 test._calls.append({'exception': e})
#                 return [e, e]
#             result = pipe.execute(raise_on_error=False)
#             test._calls.append({'result': result})
#             return result
#
#         pipe = r.pipeline(transaction=False)
#         orig_perform_execute_pipeline = pipe.perform_execute_pipeline
#         pipe.perform_execute_pipeline = perform_execute_pipeline
#
#         try:
#             pipe.set('foo', 1)
#             pipe.get('foo')
#             res = pipe.execute()
#             assert res == [True, b'1']
#             assert isinstance(test._calls[0]['exception'], ResponseError)
#             assert re.match("ASK", str(test._calls[0]['exception']))
#             assert isinstance(test._calls[1]['result'][0], ResponseError)
#             assert re.match("MOVED", str(test._calls[1]['result'][0]))
#             assert test._calls[2] == {'result': [True, b'1']}
#         finally:
#             pipe.perform_execute_pipeline = orig_perform_execute_pipeline
#             del test._calls
#
#     def test_empty_stack(self, r):
#         """
#         If pipeline is executed with no commands it should
#         return a empty list.
#         """
#         p = r.pipeline()
#         result = p.execute()
#         assert result == []
#
#
# class TestReadOnlyPipeline:
#
#     def test_pipeline_readonly(self, r, ro):
#         """
#         On readonly mode, we supports get related stuff only.
#         """
#         r.set('foo71', 'a1')   # we assume this key is set on 127.0.0.1:7001
#         r.zadd('foo88', z1=1)  # we assume this key is set on 127.0.0.1:7002
#         r.zadd('foo88', z2=4)
#
#         with ro.pipeline() as readonly_pipe:
#             readonly_pipe.get('foo71').zrange('foo88', 0, 5, withscores=True)
#             assert readonly_pipe.execute() == [
#                 b('a1'),
#                 [(b('z1'), 1.0), (b('z2'), 4)],
#             ]
#
#     def assert_moved_redirection_on_slave(self, connection_pool_cls, cluster_obj):
#         with patch.object(connection_pool_cls, 'get_node_by_slot') as return_slave_mock:
#             with patch.object(ClusterConnectionPool, 'get_master_node_by_slot') as return_master_mock:
#                 def get_mock_node(role, port):
#                     return {
#                         'name': '127.0.0.1:{0}'.format(port),
#                         'host': '127.0.0.1',
#                         'port': port,
#                         'server_type': role,
#                     }
#
#                 return_slave_mock.return_value = get_mock_node('slave', 7005)
#                 return_master_mock.return_value = get_mock_node('slave', 7001)
#
#                 with cluster_obj.pipeline() as pipe:
#                     # we assume this key is set on 127.0.0.1:7001(7004)
#                     pipe.get('foo87').get('foo88').execute() == [None, None]
#                     assert return_master_mock.call_count == 2
#
#     def test_moved_redirection_on_slave_with_default(self):
#         """
#         On Pipeline, we redirected once and finally get from master with
#         readonly client when data is completely moved.
#         """
#         self.assert_moved_redirection_on_slave(
#             ClusterConnectionPool,
#             StrictRedisCluster(host="127.0.0.1", port=7000, reinitialize_steps=1)
#         )
#
#     def test_moved_redirection_on_slave_with_readonly_mode_client(self):
#         """
#         Ditto with READONLY mode.
#         """
#         self.assert_moved_redirection_on_slave(
#             ClusterReadOnlyConnectionPool,
#             StrictRedisCluster(host="127.0.0.1", port=7000, readonly_mode=True, reinitialize_steps=1)
#         )
#
#     def test_access_correct_slave_with_readonly_mode_client(self, sr):
#         """
#         Test that the client can get value normally with readonly mode
#         when we connect to correct slave.
#         """
#
#         # we assume this key is set on 127.0.0.1:7001
#         sr.set('foo87', 'foo')
#         sr.set('foo88', 'bar')
#         import time
#         time.sleep(1)
#
#         with patch.object(ClusterReadOnlyConnectionPool, 'get_node_by_slot') as return_slave_mock:
#             return_slave_mock.return_value = {
#                 'name': '127.0.0.1:7004',
#                 'host': '127.0.0.1',
#                 'port': 7004,
#                 'server_type': 'slave',
#             }
#
#             master_value = {'host': '127.0.0.1', 'name': '127.0.0.1:7001', 'port': 7001, 'server_type': 'master'}
#             with patch.object(
#                     ClusterConnectionPool,
#                     'get_master_node_by_slot',
#                     return_value=master_value) as return_master_mock:
#                 readonly_client = StrictRedisCluster(host="127.0.0.1", port=7000, readonly_mode=True)
#                 with readonly_client.pipeline() as readonly_pipe:
#                     assert readonly_pipe.get('foo88').get('foo87').execute() == [b('bar'), b('foo')]
#                     assert return_master_mock.call_count == 0



from __future__ import with_statement

import asyncio
import time

import pytest

from coredis.exceptions import LockError, ResponseError
from coredis.lock import Lock, LuaLock


class TestLock:
    lock_class = Lock

    def get_lock(self, redis, *args, **kwargs):
        kwargs["lock_class"] = self.lock_class
        return redis.lock(*args, **kwargs)

    @pytest.mark.asyncio()
    async def test_lock(self, r):
        await r.flushdb()
        lock = self.get_lock(r, "foo")
        assert await lock.acquire(blocking=False)
        assert await r.get("foo") == lock.local.get()
        assert await r.ttl("foo") == -1
        await lock.release()
        assert await r.get("foo") is None

    @pytest.mark.asyncio()
    async def test_competing_locks(self, r):
        lock1 = self.get_lock(r, "foo")
        lock2 = self.get_lock(r, "foo")
        assert await lock1.acquire(blocking=False)
        assert not await lock2.acquire(blocking=False)
        await lock1.release()
        assert await lock2.acquire(blocking=False)
        assert not await lock1.acquire(blocking=False)
        await lock2.release()

    @pytest.mark.asyncio()
    async def test_timeout(self, r):
        lock = self.get_lock(r, "foo", timeout=10)
        assert await lock.acquire(blocking=False)
        assert 8 < await r.ttl("foo") <= 10
        await lock.release()

    @pytest.mark.asyncio()
    async def test_float_timeout(self, r):
        lock = self.get_lock(r, "foo", timeout=9.5)
        assert await lock.acquire(blocking=False)
        assert 8 < await r.pttl("foo") <= 9500
        await lock.release()

    @pytest.mark.asyncio()
    async def test_blocking_timeout(self, r):
        lock1 = self.get_lock(r, "foo")
        assert await lock1.acquire(blocking=False)
        lock2 = self.get_lock(r, "foo", blocking_timeout=0.2)
        start = time.time()
        assert not await lock2.acquire()
        assert (time.time() - start) > 0.2
        await lock1.release()

    @pytest.mark.asyncio()
    async def test_context_manager(self, r):
        # blocking_timeout prevents a deadlock if the lock can't be acquired
        # for some reason
        async with self.get_lock(r, "foo", blocking_timeout=0.2) as lock:
            assert await r.get("foo") == lock.local.get()
        assert await r.get("foo") is None

    @pytest.mark.asyncio()
    async def test_high_sleep_raises_error(self, r):
        "If sleep is higher than timeout, it should raise an error"
        with pytest.raises(LockError):
            self.get_lock(r, "foo", timeout=1, sleep=2)

    @pytest.mark.asyncio()
    async def test_releasing_unlocked_lock_raises_error(self, r):
        lock = self.get_lock(r, "foo")
        with pytest.raises(LockError):
            await lock.release()

    @pytest.mark.asyncio()
    async def test_releasing_lock_no_longer_owned_raises_error(self, r):
        lock = self.get_lock(r, "foo")
        await lock.acquire(blocking=False)
        # manually change the token
        await r.set("foo", "a")
        with pytest.raises(LockError):
            await lock.release()
        # even though we errored, the token is still cleared
        assert lock.local.get() is None

    @pytest.mark.asyncio()
    async def test_extend_lock(self, r):
        await r.flushdb()
        lock = self.get_lock(r, "foo", timeout=10)
        assert await lock.acquire(blocking=False)
        assert 8000 < await r.pttl("foo") <= 10000
        assert await lock.extend(10)
        assert 16000 < await r.pttl("foo") <= 20000
        await lock.release()

    @pytest.mark.asyncio()
    async def test_extend_lock_float(self, r):
        await r.flushdb()
        lock = self.get_lock(r, "foo", timeout=10.0)
        assert await lock.acquire(blocking=False)
        assert 8000 < await r.pttl("foo") <= 10000
        assert await lock.extend(10.0)
        assert 16000 < await r.pttl("foo") <= 20000
        await lock.release()

    @pytest.mark.asyncio()
    async def test_extending_unlocked_lock_raises_error(self, r):
        lock = self.get_lock(r, "foo", timeout=10)
        with pytest.raises(LockError):
            await lock.extend(10)

    @pytest.mark.asyncio()
    async def test_extending_lock_with_no_timeout_raises_error(self, r):
        lock = self.get_lock(r, "foo")
        await r.flushdb()
        assert await lock.acquire(blocking=False)
        with pytest.raises(LockError):
            await lock.extend(10)
        await lock.release()

    @pytest.mark.asyncio()
    async def test_extending_lock_no_longer_owned_raises_error(self, r):
        lock = self.get_lock(r, "foo")
        await r.flushdb()
        assert await lock.acquire(blocking=False)
        await r.set("foo", "a")
        with pytest.raises(LockError):
            await lock.extend(10)


class TestLuaLock(TestLock):
    lock_class = LuaLock


class TestLockClassSelection:
    @pytest.mark.asyncio()
    async def test_lock_class_argument(self, r):
        lock = r.lock("foo", lock_class=Lock)
        assert type(lock) == Lock
        lock = r.lock("foo", lock_class=LuaLock)
        assert type(lock) == LuaLock

    @pytest.mark.asyncio()
    async def test_cached_lualock_flag(self, r):
        try:
            r._use_lua_lock = True
            lock = r.lock("foo")
            assert type(lock) == LuaLock
        finally:
            r._use_lua_lock = None

    @pytest.mark.asyncio()
    async def test_cached_lock_flag(self, r):
        try:
            r._use_lua_lock = False
            lock = r.lock("foo")
            assert type(lock) == Lock
        finally:
            r._use_lua_lock = None

    @pytest.mark.asyncio()
    async def test_lua_compatible_server(self, r, monkeypatch):
        @classmethod
        def mock_register(cls, redis):
            return

        monkeypatch.setattr(LuaLock, "register_scripts", mock_register)
        try:
            lock = r.lock("foo")
            assert type(lock) == LuaLock
            assert r._use_lua_lock is True
        finally:
            r._use_lua_lock = None

    @pytest.mark.asyncio()
    async def test_lua_unavailable(self, r, monkeypatch):
        @classmethod
        def mock_register(cls, redis):
            raise ResponseError()

        monkeypatch.setattr(LuaLock, "register_scripts", mock_register)
        try:
            lock = r.lock("foo")
            assert type(lock) == Lock
            assert r._use_lua_lock is False
        finally:
            r._use_lua_lock = None



# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement
import datetime
import re
import time
import pytest
from string import ascii_letters

# rediscluster imports
from coredis.exceptions import (
    RedisClusterException,
    ResponseError,
    DataError,
    RedisError,
)
from coredis.utils import b, iteritems, iterkeys, itervalues
from coredis.commands.server import parse_info


async def redis_server_time(client):
    t = await client.time()
    seconds, milliseconds = list(t.values())[0]
    timestamp = float("{0}.{1}".format(seconds, milliseconds))
    return datetime.datetime.fromtimestamp(timestamp)


class TestRedisCommands:
    @pytest.mark.asyncio
    async def test_zrevrangebylex(self, r):
        await r.flushdb()
        await r.zadd("a", a=0, b=0, c=0, d=0, e=0, f=0, g=0)
        assert await r.zrevrangebylex("a", "[c", "-") == [b("c"), b("b"), b("a")]
        assert await r.zrevrangebylex("a", "(c", "-") == [b("b"), b("a")]
        assert await r.zrevrangebylex("a", "(g", "[aaa") == [
            b("f"),
            b("e"),
            b("d"),
            b("c"),
            b("b"),
        ]
        assert await r.zrevrangebylex("a", "+", "[f") == [b("g"), b("f")]
        assert await r.zrevrangebylex("a", "+", "-", start=3, num=2) == [b("d"), b("c")]

    @pytest.mark.asyncio
    async def test_command_on_invalid_key_type(self, r):
        await r.flushdb()
        await r.lpush("a", "1")
        with pytest.raises(ResponseError):
            await r.get("a")

    # SERVER INFORMATION
    @pytest.mark.asyncio
    async def test_client_list(self, r):
        client_lists = await r.client_list()
        for server, clients in client_lists.items():
            assert isinstance(clients[0], dict)
            assert "addr" in clients[0]

    @pytest.mark.asyncio
    async def test_client_getname(self, r):
        client_names = await r.client_getname()
        for server, name in client_names.items():
            assert name is None

    @pytest.mark.asyncio
    async def test_client_setname(self, r):
        with pytest.raises(RedisClusterException):
            assert await r.client_setname("redis_py_test")

    @pytest.mark.asyncio
    async def test_config_get(self, r):
        config = await r.config_get()
        for server, data in config.items():
            assert "maxmemory" in data
            assert data["maxmemory"].isdigit()

    @pytest.mark.asyncio
    async def test_config_resetstat(self, r):
        await r.ping()
        info = await r.info()
        for server, info in info.items():
            prior_commands_processed = int(info["total_commands_processed"])
            assert prior_commands_processed >= 1
        await r.config_resetstat()
        info = await r.info()
        for server, info in info.items():
            reset_commands_processed = int(info["total_commands_processed"])
            assert reset_commands_processed < prior_commands_processed

    @pytest.mark.asyncio
    async def test_config_set(self, r):
        assert await r.config_set("dbfilename", "redis_py_test.rdb")
        config = await r.config_get()
        for server, config in config.items():
            assert config["dbfilename"] == "redis_py_test.rdb"

    @pytest.mark.asyncio
    async def test_echo(self, r):
        echo = await r.echo("foo bar")
        for server, res in echo.items():
            assert res == b("foo bar")

    @pytest.mark.asyncio
    async def test_object(self, r):
        await r.flushdb()
        await r.set("a", "foo")
        assert await r.object("refcount", "a") == 1
        assert isinstance(await r.object("refcount", "a"), int)
        # assert isinstance(await r.object('idletime', 'a'), int)
        # assert await r.object('encoding', 'a') in (b('raw'), b('embstr'))
        assert await r.object("idletime", "invalid-key") is None

    @pytest.mark.asyncio
    async def test_ping(self, r):
        assert await r.ping()

    @pytest.mark.asyncio
    async def test_time(self, r):
        await r.flushdb()
        time = await r.time()
        for t in time.values():
            assert len(t) == 2
            assert isinstance(t[0], int)
            assert isinstance(t[1], int)

    # BASIC KEY COMMANDS
    @pytest.mark.asyncio
    async def test_append(self, r):
        await r.flushdb()
        assert await r.append("a", "a1") == 2
        assert await r.get("a") == b("a1")
        assert await r.append("a", "a2") == 4
        assert await r.get("a") == b("a1a2")

    @pytest.mark.asyncio
    async def test_bitcount(self, r):
        await r.flushdb()
        await r.setbit("a", 5, True)
        assert await r.bitcount("a") == 1
        await r.setbit("a", 6, True)
        assert await r.bitcount("a") == 2
        await r.setbit("a", 5, False)
        assert await r.bitcount("a") == 1
        await r.setbit("a", 9, True)
        await r.setbit("a", 17, True)
        await r.setbit("a", 25, True)
        await r.setbit("a", 33, True)
        assert await r.bitcount("a") == 5
        assert await r.bitcount("a", 0, -1) == 5
        assert await r.bitcount("a", 2, 3) == 2
        assert await r.bitcount("a", 2, -1) == 3
        assert await r.bitcount("a", -2, -1) == 2
        assert await r.bitcount("a", 1, 1) == 1

    @pytest.mark.asyncio
    async def test_bitop_not_supported(self, r):
        await r.flushdb()
        await r.set("a", "")
        with pytest.raises(RedisClusterException):
            await r.bitop("not", "r", "a")

    @pytest.mark.asyncio
    async def test_bitpos(self, r):
        """
        Bitpos was added in redis-py in version 2.10.2

        # TODO: Added b() around keys but i think they should not have to be
                there for this command to work properly.
        """
        await r.flushdb()
        key = "key:bitpos"
        await r.set(key, b("\xff\xf0\x00"))
        assert await r.bitpos(key, 0) == 12
        assert await r.bitpos(key, 0, 2, -1) == 16
        assert await r.bitpos(key, 0, -2, -1) == 12
        await r.set(key, b("\x00\xff\xf0"))
        assert await r.bitpos(key, 1, 0) == 8
        assert await r.bitpos(key, 1, 1) == 8
        await r.set(key, "\x00\x00\x00")
        assert await r.bitpos(key, 1) == -1

    @pytest.mark.asyncio
    async def test_bitpos_wrong_arguments(self, r):
        """
        Bitpos was added in redis-py in version 2.10.2
        """
        key = "key:bitpos:wrong:args"
        await r.flushdb()
        await r.set(key, b("\xff\xf0\x00"))
        with pytest.raises(RedisError):
            await r.bitpos(key, 0, end=1) == 12
        with pytest.raises(RedisError):
            await r.bitpos(key, 7) == 12

    @pytest.mark.asyncio
    async def test_decr(self, r):
        await r.flushdb()
        assert await r.decr("a") == -1
        assert await r.get("a") == b("-1")
        assert await r.decr("a") == -2
        assert await r.get("a") == b("-2")
        assert await r.decr("a", amount=5) == -7
        assert await r.get("a") == b("-7")

    @pytest.mark.asyncio
    async def test_delete(self, r):
        await r.flushdb()
        assert await r.delete("a") == 0
        await r.set("a", "foo")
        assert await r.delete("a") == 1

    @pytest.mark.asyncio
    async def test_delete_with_multiple_keys(self, r):
        await r.flushdb()
        await r.set("a", "foo")
        await r.set("b", "bar")
        assert await r.delete("a", "b") == 2
        assert await r.get("a") is None
        assert await r.get("b") is None

    @pytest.mark.asyncio
    async def test_delitem(self, r):
        await r.flushdb()
        await r.set("a", "foo")
        await r.delete("a")
        assert await r.get("a") is None

    @pytest.mark.asyncio
    async def test_dump_and_restore(self, r):
        await r.flushdb()
        await r.set("a", "foo")
        dumped = await r.dump("a")
        await r.delete("a")
        await r.restore("a", 0, dumped)
        assert await r.get("a") == b("foo")

    @pytest.mark.asyncio
    async def test_exists(self, r):
        await r.flushdb()
        assert not await r.exists("a")
        await r.set("a", "foo")
        assert await r.exists("a")

    @pytest.mark.asyncio
    async def test_exists_contains(self, r):
        await r.flushdb()
        assert not await r.exists("a")
        await r.set("a", "foo")
        assert await r.exists("a")

    @pytest.mark.asyncio
    async def test_expire(self, r):
        await r.flushdb()
        assert not await r.expire("a", 10)
        await r.set("a", "foo")
        assert await r.expire("a", 10)
        assert 0 < await r.ttl("a") <= 10
        assert await r.persist("a")
        # the ttl command changes behavior in redis-2.8+ http://redis.io/commands/ttl
        assert await r.ttl("a") == -1

    @pytest.mark.asyncio
    async def test_expireat_datetime(self, r):
        await r.flushdb()
        expire_at = await redis_server_time(r) + datetime.timedelta(minutes=1)
        await r.set("a", "foo")
        assert await r.expireat("a", expire_at)
        assert 0 < await r.ttl("a") <= 61

    @pytest.mark.asyncio
    async def test_expireat_no_key(self, r):
        await r.flushdb()
        expire_at = await redis_server_time(r) + datetime.timedelta(minutes=1)
        assert not await r.expireat("a", expire_at)

    @pytest.mark.asyncio
    async def test_expireat_unixtime(self, r):
        expire_at = await redis_server_time(r) + datetime.timedelta(minutes=1)
        await r.set("a", "foo")
        expire_at_seconds = int(time.mktime(expire_at.timetuple()))
        assert await r.expireat("a", expire_at_seconds)
        assert 0 < await r.ttl("a") <= 61

    @pytest.mark.asyncio
    async def test_get_and_set(self, r):
        # get and set can't be tested independently of each other
        await r.flushdb()
        assert await r.get("a") is None
        byte_string = b("value")
        integer = 5
        unicode_string = str(3456) + "abcd" + str(3421)
        assert await r.set("byte_string", byte_string)
        assert await r.set("integer", 5)
        assert await r.set("unicode_string", unicode_string)
        assert await r.get("byte_string") == byte_string
        assert await r.get("integer") == b(str(integer))
        assert (await r.get("unicode_string")).decode("utf-8") == unicode_string

    @pytest.mark.asyncio
    async def test_getitem_and_setitem(self, r):
        await r.flushdb()
        await r.set("a", "bar")
        assert await r.get("a") == b("bar")

    @pytest.mark.asyncio
    async def test_get_set_bit(self, r):
        await r.flushdb()
        # no value
        assert not await r.getbit("a", 5)
        # set bit 5
        assert not await r.setbit("a", 5, True)
        assert await r.getbit("a", 5)
        # unset bit 4
        assert not await r.setbit("a", 4, False)
        assert not await r.getbit("a", 4)
        # set bit 4
        assert not await r.setbit("a", 4, True)
        assert await r.getbit("a", 4)
        # set bit 5 again
        assert await r.setbit("a", 5, True)
        assert await r.getbit("a", 5)

    @pytest.mark.asyncio
    async def test_getrange(self, r):
        await r.flushdb()
        await r.set("a", "foo")
        assert await r.getrange("a", 0, 0) == b("f")
        assert await r.getrange("a", 0, 2) == b("foo")
        assert await r.getrange("a", 3, 4) == b("")

    @pytest.mark.asyncio
    async def test_getset(self, r):
        await r.flushdb()
        assert await r.getset("a", "foo") is None
        assert await r.getset("a", "bar") == b("foo")
        assert await r.get("a") == b("bar")

    @pytest.mark.asyncio
    async def test_incr(self, r):
        await r.flushdb()
        assert await r.incr("a") == 1
        assert await r.get("a") == b("1")
        assert await r.incr("a") == 2
        assert await r.get("a") == b("2")
        assert await r.incr("a", amount=5) == 7
        assert await r.get("a") == b("7")

    @pytest.mark.asyncio
    async def test_incrby(self, r):
        await r.flushdb()
        assert await r.incrby("a") == 1
        assert await r.incrby("a", 4) == 5
        assert await r.get("a") == b("5")

    @pytest.mark.asyncio
    async def test_incrbyfloat(self, r):
        await r.flushdb()
        assert await r.incrbyfloat("a") == 1.0
        assert await r.get("a") == b("1")
        assert await r.incrbyfloat("a", 1.1) == 2.1
        assert float(await r.get("a")) == float(2.1)

    @pytest.mark.asyncio
    async def test_keys(self, r):
        await r.flushdb()
        keys = await r.keys()
        assert keys == []
        keys_with_underscores = set(["test_a", "test_b"])
        keys = keys_with_underscores.union(set(["testc"]))
        for key in keys:
            await r.set(key, 1)
        assert set(await r.keys(pattern="test_*")) == {
            b(k) for k in keys_with_underscores
        }
        assert set(await r.keys(pattern="test*")) == {b(k) for k in keys}

    @pytest.mark.asyncio
    async def test_mget(self, r):
        await r.flushdb()
        assert await r.mget(["a", "b"]) == [None, None]
        await r.set("a", 1)
        await r.set("b", 2)
        await r.set("c", 3)
        assert await r.mget("a", "other", "b", "c") == [b("1"), None, b("2"), b("3")]

    @pytest.mark.asyncio
    async def test_mset(self, r):
        await r.flushdb()
        d = {"a": b("1"), "b": b("2"), "c": b("3")}
        assert await r.mset(d)
        for k, v in iteritems(d):
            assert await r.mget(k) == [v]

    @pytest.mark.asyncio
    async def test_mset_kwargs(self, r):
        await r.flushdb()
        d = {"a": b("1"), "b": b("2"), "c": b("3")}
        assert await r.mset(**d)
        for k, v in iteritems(d):
            assert await r.get(k) == v

    @pytest.mark.asyncio
    async def test_msetnx(self, r):
        await r.flushdb()
        d = {"a": b("1"), "b": b("2"), "c": b("3")}
        assert await r.msetnx(d)
        d2 = {"a": b("x"), "d": b("4")}
        assert not await r.msetnx(d2)
        for k, v in iteritems(d):
            assert await r.get(k) == v
        assert await r.get("d") is None

    @pytest.mark.asyncio
    async def test_msetnx_kwargs(self, r):
        await r.flushdb()
        d = {"a": b("1"), "b": b("2"), "c": b("3")}
        assert await r.msetnx(**d)
        d2 = {"a": b("x"), "d": b("4")}
        assert not await r.msetnx(**d2)
        for k, v in iteritems(d):
            assert await r.get(k) == v
        assert await r.get("d") is None

    @pytest.mark.asyncio
    async def test_pexpire(self, r):
        await r.flushdb()
        assert not await r.pexpire("a", 60000)
        await r.set("a", "foo")
        assert await r.pexpire("a", 60000)
        assert 0 < await r.pttl("a") <= 60000
        assert await r.persist("a")
        # redis-py tests seemed to be for older version of redis?
        # redis-2.8+ returns -1 if key exists but is non-expiring: http://redis.io/commands/pttl
        assert await r.pttl("a") == -1

    @pytest.mark.asyncio
    async def test_pexpireat_datetime(self, r):
        await r.flushdb()
        expire_at = await redis_server_time(r) + datetime.timedelta(minutes=1)
        await r.set("a", "foo")
        assert await r.pexpireat("a", expire_at)
        assert 0 < await r.pttl("a") <= 61000

    @pytest.mark.asyncio
    async def test_pexpireat_no_key(self, r):
        await r.flushdb()
        expire_at = await redis_server_time(r) + datetime.timedelta(minutes=1)
        assert not await r.pexpireat("a", expire_at)

    @pytest.mark.asyncio
    async def test_pexpireat_unixtime(self, r):
        await r.flushdb()
        expire_at = await redis_server_time(r) + datetime.timedelta(minutes=1)
        await r.set("a", "foo")
        expire_at_seconds = int(time.mktime(expire_at.timetuple())) * 1000
        assert await r.pexpireat("a", expire_at_seconds)
        assert 0 < await r.pttl("a") <= 61000

    @pytest.mark.asyncio
    async def test_psetex(self, r):
        await r.flushdb()
        assert await r.psetex("a", 1000, "value")
        assert await r.get("a") == b("value")
        assert 0 < await r.pttl("a") <= 1000

    @pytest.mark.asyncio
    async def test_psetex_timedelta(self, r):
        await r.flushdb()
        expire_at = datetime.timedelta(milliseconds=1000)
        assert await r.psetex("a", expire_at, "value")
        assert await r.get("a") == b("value")
        assert 0 < await r.pttl("a") <= 1000

    @pytest.mark.asyncio
    async def test_randomkey(self, r):
        await r.flushdb()
        assert await r.randomkey() is None
        for key in ("a", "b", "c"):
            await r.set(key, 1)
        assert await r.randomkey() in (b("a"), b("b"), b("c"))

    @pytest.mark.asyncio
    async def test_rename(self, r):
        await r.flushdb()
        await r.set("a", "1")
        assert await r.rename("a", "b")
        assert await r.get("a") is None
        assert await r.get("b") == b("1")

        with pytest.raises(ResponseError) as ex:
            await r.rename("foo", "foo")
        assert str(ex.value).startswith("source and destination objects are the same")

        assert await r.get("foo") is None
        with pytest.raises(ResponseError) as ex:
            await r.rename("foo", "bar")
        assert str(ex.value).startswith("no such key")

    @pytest.mark.asyncio
    async def test_renamenx(self, r):
        await r.flushdb()
        await r.set("a", "1")
        await r.set("b", "2")
        assert not await r.renamenx("a", "b")
        assert await r.get("a") == b("1")
        assert await r.get("b") == b("2")

        assert await r.renamenx("a", "c")
        assert await r.get("c") == b("1")

    @pytest.mark.asyncio
    async def test_set_nx(self, r):
        await r.flushdb()
        assert await r.set("a", "1", nx=True)
        assert not await r.set("a", "2", nx=True)
        assert await r.get("a") == b("1")

    @pytest.mark.asyncio
    async def test_set_xx(self, r):
        await r.flushdb()
        assert not await r.set("a", "1", xx=True)
        assert await r.get("a") is None
        await r.set("a", "bar")
        assert await r.set("a", "2", xx=True)
        assert await r.get("a") == b("2")

    @pytest.mark.asyncio
    async def test_set_px(self, r):
        await r.flushdb()
        assert await r.set("a", "1", px=10000)
        assert await r.get("a") == b("1")
        assert 0 < await r.pttl("a") <= 10000
        assert 0 < await r.ttl("a") <= 10

    @pytest.mark.asyncio
    async def test_set_px_timedelta(self, r):
        await r.flushdb()
        expire_at = datetime.timedelta(milliseconds=1000)
        assert await r.set("a", "1", px=expire_at)
        assert 0 < await r.pttl("a") <= 1000
        assert 0 < await r.ttl("a") <= 1

    @pytest.mark.asyncio
    async def test_set_ex(self, r):
        await r.flushdb()
        assert await r.set("a", "1", ex=10)
        assert 0 < await r.ttl("a") <= 10

    @pytest.mark.asyncio
    async def test_set_ex_timedelta(self, r):
        await r.flushdb()
        expire_at = datetime.timedelta(seconds=60)
        assert await r.set("a", "1", ex=expire_at)
        assert 0 < await r.ttl("a") <= 60

    @pytest.mark.asyncio
    async def test_set_multipleoptions(self, r):
        await r.flushdb()
        await r.set("a", "val")
        assert await r.set("a", "1", xx=True, px=10000)
        assert 0 < await r.ttl("a") <= 10

    @pytest.mark.asyncio
    async def test_setex(self, r):
        await r.flushdb()
        assert await r.setex("a", 60, "1")
        assert await r.get("a") == b("1")
        assert 0 < await r.ttl("a") <= 60

    @pytest.mark.asyncio
    async def test_setnx(self, r):
        await r.flushdb()
        assert await r.setnx("a", "1")
        assert await r.get("a") == b("1")
        assert not await r.setnx("a", "2")
        assert await r.get("a") == b("1")

    @pytest.mark.asyncio
    async def test_setrange(self, r):
        await r.flushdb()
        assert await r.setrange("a", 5, "foo") == 8
        assert await r.get("a") == b("\0\0\0\0\0foo")
        await r.set("a", "abcasync defghijh")
        assert await r.setrange("a", 6, "12345") == 17
        assert await r.get("a") == b("abcasy12345fghijh")

    @pytest.mark.asyncio
    async def test_strlen(self, r):
        await r.flushdb()
        await r.set("a", "foo")
        assert await r.strlen("a") == 3

    @pytest.mark.asyncio
    async def test_substr(self, r):
        await r.flushdb()
        await r.set("a", "0123456789")
        assert await r.substr("a", 0) == b("0123456789")
        assert await r.substr("a", 2) == b("23456789")
        assert await r.substr("a", 3, 5) == b("345")
        assert await r.substr("a", 3, -2) == b("345678")

    @pytest.mark.asyncio
    async def test_type(self, r):
        await r.flushdb()
        assert await r.type("a") == b("none")
        await r.set("a", "1")
        assert await r.type("a") == b("string")
        await r.delete("a")
        await r.lpush("a", "1")
        assert await r.type("a") == b("list")
        await r.delete("a")
        await r.sadd("a", "1")
        assert await r.type("a") == b("set")
        await r.delete("a")
        await r.zadd("a", **{"1": 1})
        assert await r.type("a") == b("zset")

    # LIST COMMANDS
    @pytest.mark.asyncio
    async def test_blpop(self, r):
        await r.flushdb()
        await r.rpush("a{foo}", "1", "2")
        await r.rpush("b{foo}", "3", "4")
        assert await r.blpop(["b{foo}", "a{foo}"], timeout=1) == (b("b{foo}"), b("3"))
        assert await r.blpop(["b{foo}", "a{foo}"], timeout=1) == (b("b{foo}"), b("4"))
        assert await r.blpop(["b{foo}", "a{foo}"], timeout=1) == (b("a{foo}"), b("1"))
        assert await r.blpop(["b{foo}", "a{foo}"], timeout=1) == (b("a{foo}"), b("2"))
        assert await r.blpop(["b{foo}", "a{foo}"], timeout=1) is None
        await r.rpush("c{foo}", "1")
        assert await r.blpop("c{foo}", timeout=1) == (b("c{foo}"), b("1"))

    @pytest.mark.asyncio
    async def test_brpop(self, r):
        await r.flushdb()
        await r.rpush("a{foo}", "1", "2")
        await r.rpush("b{foo}", "3", "4")
        assert await r.brpop(["b{foo}", "a{foo}"], timeout=1) == (b("b{foo}"), b("4"))
        assert await r.brpop(["b{foo}", "a{foo}"], timeout=1) == (b("b{foo}"), b("3"))
        assert await r.brpop(["b{foo}", "a{foo}"], timeout=1) == (b("a{foo}"), b("2"))
        assert await r.brpop(["b{foo}", "a{foo}"], timeout=1) == (b("a{foo}"), b("1"))
        assert await r.brpop(["b{foo}", "a{foo}"], timeout=1) is None
        await r.rpush("c{foo}", "1")
        assert await r.brpop("c{foo}", timeout=1) == (b("c{foo}"), b("1"))

    @pytest.mark.asyncio
    async def test_brpoplpush(self, r):
        await r.flushdb()
        await r.rpush("a{foo}", "1", "2")
        await r.rpush("b{foo}", "3", "4")
        assert await r.brpoplpush("a{foo}", "b{foo}") == b("2")
        assert await r.brpoplpush("a{foo}", "b{foo}") == b("1")
        assert await r.brpoplpush("a{foo}", "b{foo}", timeout=1) is None
        assert await r.lrange("a{foo}", 0, -1) == []
        assert await r.lrange("b{foo}", 0, -1) == [b("1"), b("2"), b("3"), b("4")]

    @pytest.mark.asyncio
    async def test_brpoplpush_empty_string(self, r):
        await r.flushdb()
        await r.rpush("a", "")
        assert await r.brpoplpush("a", "b") == b("")

    @pytest.mark.asyncio
    async def test_lindex(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3")
        assert await r.lindex("a", "0") == b("1")
        assert await r.lindex("a", "1") == b("2")
        assert await r.lindex("a", "2") == b("3")

    @pytest.mark.asyncio
    async def test_linsert(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3")
        assert await r.linsert("a", "after", "2", "2.5") == 4
        assert await r.lrange("a", 0, -1) == [b("1"), b("2"), b("2.5"), b("3")]
        assert await r.linsert("a", "before", "2", "1.5") == 5
        assert await r.lrange("a", 0, -1) == [
            b("1"),
            b("1.5"),
            b("2"),
            b("2.5"),
            b("3"),
        ]

    @pytest.mark.asyncio
    async def test_llen(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3")
        assert await r.llen("a") == 3

    @pytest.mark.asyncio
    async def test_lpop(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3")
        assert await r.lpop("a") == b("1")
        assert await r.lpop("a") == b("2")
        assert await r.lpop("a") == b("3")
        assert await r.lpop("a") is None

    @pytest.mark.asyncio
    async def test_lpush(self, r):
        await r.flushdb()
        assert await r.lpush("a", "1") == 1
        assert await r.lpush("a", "2") == 2
        assert await r.lpush("a", "3", "4") == 4
        assert await r.lrange("a", 0, -1) == [b("4"), b("3"), b("2"), b("1")]

    @pytest.mark.asyncio
    async def test_lpushx(self, r):
        await r.flushdb()
        assert await r.lpushx("a", "1") == 0
        assert await r.lrange("a", 0, -1) == []
        await r.rpush("a", "1", "2", "3")
        assert await r.lpushx("a", "4") == 4
        assert await r.lrange("a", 0, -1) == [b("4"), b("1"), b("2"), b("3")]

    @pytest.mark.asyncio
    async def test_lrange(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3", "4", "5")
        assert await r.lrange("a", 0, 2) == [b("1"), b("2"), b("3")]
        assert await r.lrange("a", 2, 10) == [b("3"), b("4"), b("5")]
        assert await r.lrange("a", 0, -1) == [b("1"), b("2"), b("3"), b("4"), b("5")]

    @pytest.mark.asyncio
    async def test_lrem(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "1", "1", "1")
        assert await r.lrem("a", "1", 1) == 1
        assert await r.lrange("a", 0, -1) == [b("1"), b("1"), b("1")]
        assert await r.lrem("a", 0, "1") == 3
        assert await r.lrange("a", 0, -1) == []

    @pytest.mark.asyncio
    async def test_lset(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3")
        assert await r.lrange("a", 0, -1) == [b("1"), b("2"), b("3")]
        assert await r.lset("a", 1, "4")
        assert await r.lrange("a", 0, 2) == [b("1"), b("4"), b("3")]

    @pytest.mark.asyncio
    async def test_ltrim(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3")
        assert await r.ltrim("a", 0, 1)
        assert await r.lrange("a", 0, -1) == [b("1"), b("2")]

    @pytest.mark.asyncio
    async def test_rpop(self, r):
        await r.flushdb()
        await r.rpush("a", "1", "2", "3")
        assert await r.rpop("a") == b("3")
        assert await r.rpop("a") == b("2")
        assert await r.rpop("a") == b("1")
        assert await r.rpop("a") is None

    @pytest.mark.asyncio
    async def test_rpoplpush(self, r):
        await r.flushdb()
        await r.rpush("a", "a1", "a2", "a3")
        await r.rpush("b", "b1", "b2", "b3")
        assert await r.rpoplpush("a", "b") == b("a3")
        assert await r.lrange("a", 0, -1) == [b("a1"), b("a2")]
        assert await r.lrange("b", 0, -1) == [b("a3"), b("b1"), b("b2"), b("b3")]

    @pytest.mark.asyncio
    async def test_rpush(self, r):
        await r.flushdb()
        assert await r.rpush("a", "1") == 1
        assert await r.rpush("a", "2") == 2
        assert await r.rpush("a", "3", "4") == 4
        assert await r.lrange("a", 0, -1) == [b("1"), b("2"), b("3"), b("4")]

    @pytest.mark.asyncio
    async def test_rpushx(self, r):
        await r.flushdb()
        assert await r.rpushx("a", "b") == 0
        assert await r.lrange("a", 0, -1) == []
        await r.rpush("a", "1", "2", "3")
        assert await r.rpushx("a", "4") == 4
        assert await r.lrange("a", 0, -1) == [b("1"), b("2"), b("3"), b("4")]

    # SCAN COMMANDS
    @pytest.mark.asyncio
    async def test_scan(self, r):
        await r.flushdb()
        await r.set("a", 1)
        await r.set("b", 2)
        await r.set("c", 3)
        keys = []
        for result in (await r.scan()).values():
            cursor, partial_keys = result
            assert cursor == 0
            keys += partial_keys

        assert set(keys) == set([b("a"), b("b"), b("c")])

        keys = []
        for result in (await r.scan(match="a")).values():
            cursor, partial_keys = result
            assert cursor == 0
            keys += partial_keys
        assert set(keys) == set([b("a")])

    @pytest.mark.asyncio
    async def test_sscan(self, r):
        await r.flushdb()
        await r.sadd("a", 1, 2, 3)
        cursor, members = await r.sscan("a")
        assert cursor == 0
        assert set(members) == set([b("1"), b("2"), b("3")])
        _, members = await r.sscan("a", match=b("1"))
        assert set(members) == set([b("1")])

    @pytest.mark.asyncio
    async def test_hscan(self, r):
        await r.flushdb()
        await r.hmset("a", {"a": 1, "b": 2, "c": 3})
        cursor, dic = await r.hscan("a")
        assert cursor == 0
        assert dic == {b("a"): b("1"), b("b"): b("2"), b("c"): b("3")}
        _, dic = await r.hscan("a", match="a")
        assert dic == {b("a"): b("1")}

    @pytest.mark.asyncio
    async def test_zscan(self, r):
        await r.flushdb()
        await r.zadd("a", 1, "a", 2, "b", 3, "c")
        cursor, pairs = await r.zscan("a")
        assert cursor == 0
        assert set(pairs) == set([(b("a"), 1), (b("b"), 2), (b("c"), 3)])
        _, pairs = await r.zscan("a", match="a")
        assert set(pairs) == set([(b("a"), 1)])

    # SET COMMANDS
    @pytest.mark.asyncio
    async def test_sadd(self, r):
        await r.flushdb()
        members = set([b("1"), b("2"), b("3")])
        await r.sadd("a", *members)
        assert await r.smembers("a") == members

    @pytest.mark.asyncio
    async def test_scard(self, r):
        await r.flushdb()
        await r.sadd("a", "1", "2", "3")
        assert await r.scard("a") == 3

    @pytest.mark.asyncio
    async def test_sdiff(self, r):
        await r.flushdb()
        await r.sadd("a{foo}", "1", "2", "3")
        assert await r.sdiff("a{foo}", "b{foo}") == set([b("1"), b("2"), b("3")])
        await r.sadd("b{foo}", "2", "3")
        assert await r.sdiff("a{foo}", "b{foo}") == set([b("1")])

    @pytest.mark.asyncio
    async def test_sdiffstore(self, r):
        await r.flushdb()
        await r.sadd("a{foo}", "1", "2", "3")
        assert await r.sdiffstore("c{foo}", "a{foo}", "b{foo}") == 3
        assert await r.smembers("c{foo}") == set([b("1"), b("2"), b("3")])
        await r.sadd("b{foo}", "2", "3")
        assert await r.sdiffstore("c{foo}", "a{foo}", "b{foo}") == 1
        assert await r.smembers("c{foo}") == set([b("1")])

        # Diff:s that return empty set should not fail
        await r.sdiffstore("d{foo}", "e{foo}") == 0

    @pytest.mark.asyncio
    async def test_sinter(self, r):
        await r.flushdb()
        await r.sadd("a{foo}", "1", "2", "3")
        assert await r.sinter("a{foo}", "b{foo}") == set()
        await r.sadd("b{foo}", "2", "3")
        assert await r.sinter("a{foo}", "b{foo}") == set([b("2"), b("3")])

    @pytest.mark.asyncio
    async def test_sinterstore(self, r):
        await r.flushdb()
        await r.sadd("a{foo}", "1", "2", "3")
        assert await r.sinterstore("c{foo}", "a{foo}", "b{foo}") == 0
        assert await r.smembers("c{foo}") == set()
        await r.sadd("b{foo}", "2", "3")
        assert await r.sinterstore("c{foo}", "a{foo}", "b{foo}") == 2
        assert await r.smembers("c{foo}") == set([b("2"), b("3")])

    @pytest.mark.asyncio
    async def test_sismember(self, r):
        await r.flushdb()
        await r.sadd("a", "1", "2", "3")
        assert await r.sismember("a", "1")
        assert await r.sismember("a", "2")
        assert await r.sismember("a", "3")
        assert not await r.sismember("a", "4")

    @pytest.mark.asyncio
    async def test_smembers(self, r):
        await r.flushdb()
        await r.sadd("a", "1", "2", "3")
        assert await r.smembers("a") == set([b("1"), b("2"), b("3")])

    @pytest.mark.asyncio
    async def test_smove(self, r):
        await r.flushdb()
        await r.sadd("a{foo}", "a1", "a2")
        await r.sadd("b{foo}", "b1", "b2")
        assert await r.smove("a{foo}", "b{foo}", "a1")
        assert await r.smembers("a{foo}") == set([b("a2")])
        assert await r.smembers("b{foo}") == set([b("b1"), b("b2"), b("a1")])

    @pytest.mark.asyncio
    async def test_spop(self, r):
        await r.flushdb()
        s = [b("1"), b("2"), b("3")]
        await r.sadd("a", *s)
        value = await r.spop("a")
        assert value in s
        assert await r.smembers("a") == set(s) - set([value])

    @pytest.mark.asyncio
    async def test_srandmember(self, r):
        await r.flushdb()
        s = [b("1"), b("2"), b("3")]
        await r.sadd("a", *s)
        assert await r.srandmember("a") in s

    @pytest.mark.asyncio
    async def test_srandmember_multi_value(self, r):
        await r.flushdb()
        s = [b("1"), b("2"), b("3")]
        await r.sadd("a", *s)
        randoms = await r.srandmember("a", number=2)
        assert len(randoms) == 2
        assert set(randoms).intersection(s) == set(randoms)

    @pytest.mark.asyncio
    async def test_srem(self, r):
        await r.flushdb()
        await r.sadd("a", "1", "2", "3", "4")
        assert await r.srem("a", "5") == 0
        assert await r.srem("a", "2", "4") == 2
        assert await r.smembers("a") == set([b("1"), b("3")])

    @pytest.mark.asyncio
    async def test_sunion(self, r):
        await r.flushdb()
        await r.sadd("a{foo}", "1", "2")
        await r.sadd("b{foo}", "2", "3")
        assert await r.sunion("a{foo}", "b{foo}") == set([b("1"), b("2"), b("3")])

    @pytest.mark.asyncio
    async def test_sunionstore(self, r):
        await r.flushdb()
        await r.sadd("a{foo}", "1", "2")
        await r.sadd("b{foo}", "2", "3")
        assert await r.sunionstore("c{foo}", "a{foo}", "b{foo}") == 3
        assert await r.smembers("c{foo}") == set([b("1"), b("2"), b("3")])

    # SORTED SET COMMANDS
    @pytest.mark.asyncio
    async def test_zadd(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zrange("a", 0, -1) == [b("a1"), b("a2"), b("a3")]

    @pytest.mark.asyncio
    async def test_zaddoption(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1)
        assert int(await r.zscore("a", "a1")) == 1
        assert int(await r.zaddoption("a", "NX", a1=2)) == 0
        assert int(await r.zaddoption("a", "NX CH", a1=2)) == 0
        assert int(await r.zscore("a", "a1")) == 1
        assert await r.zcard("a") == 1
        assert int(await r.zaddoption("a", "XX", a2=1)) == 0
        assert await r.zcard("a") == 1
        assert int(await r.zaddoption("a", "XX", a1=2)) == 0
        assert int(await r.zaddoption("a", "XX CH", a1=3)) == 1
        assert int(await r.zscore("a", "a1")) == 3
        assert int(await r.zaddoption("a", "NX", a2=1)) == 1
        assert int(await r.zaddoption("a", "NX CH", a3=1)) == 1
        assert await r.zcard("a") == 3
        await r.zaddoption("a", "INCR", a3=1)
        assert int(await r.zscore("a", "a3")) == 2

    @pytest.mark.asyncio
    async def test_zcard(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zcard("a") == 3

    @pytest.mark.asyncio
    async def test_zcount(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zcount("a", "-inf", "+inf") == 3
        assert await r.zcount("a", 1, 2) == 2
        assert await r.zcount("a", 10, 20) == 0

    @pytest.mark.asyncio
    async def test_zincrby(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zincrby("a", "a2") == 3.0
        assert await r.zincrby("a", "a3", amount=5) == 8.0
        assert await r.zscore("a", "a2") == 3.0
        assert await r.zscore("a", "a3") == 8.0

    @pytest.mark.asyncio
    async def test_zlexcount(self, r):
        await r.flushdb()
        await r.zadd("a", a=0, b=0, c=0, d=0, e=0, f=0, g=0)
        assert await r.zlexcount("a", "-", "+") == 7
        assert await r.zlexcount("a", "[b", "[f") == 5

    @pytest.mark.asyncio
    async def test_zinterstore_fail_cross_slot(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=1, a3=1)
        await r.zadd("b", a1=2, a2=2, a3=2)
        await r.zadd("c", a1=6, a3=5, a4=4)
        with pytest.raises(ResponseError) as excinfo:
            await r.zinterstore("d", ["a", "b", "c"])
        assert re.search("ClusterCrossSlotError", str(excinfo))

    @pytest.mark.asyncio
    async def test_zinterstore_sum(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=1, a3=1)
        await r.zadd("b{foo}", a1=2, a2=2, a3=2)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert await r.zinterstore("d{foo}", ["a{foo}", "b{foo}", "c{foo}"]) == 2
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a3"), 8),
            (b("a1"), 9),
        ]

    @pytest.mark.asyncio
    async def test_zinterstore_max(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=1, a3=1)
        await r.zadd("b{foo}", a1=2, a2=2, a3=2)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert (
            await r.zinterstore(
                "d{foo}", ["a{foo}", "b{foo}", "c{foo}"], aggregate="MAX"
            )
            == 2
        )
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a3"), 5),
            (b("a1"), 6),
        ]

    @pytest.mark.asyncio
    async def test_zinterstore_min(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=2, a3=3)
        await r.zadd("b{foo}", a1=2, a2=3, a3=5)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert (
            await r.zinterstore(
                "d{foo}", ["a{foo}", "b{foo}", "c{foo}"], aggregate="MIN"
            )
            == 2
        )
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a1"), 1),
            (b("a3"), 3),
        ]

    @pytest.mark.asyncio
    async def test_zinterstore_with_weight(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=1, a3=1)
        await r.zadd("b{foo}", a1=2, a2=2, a3=2)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert (
            await r.zinterstore("d{foo}", {"a{foo}": 1, "b{foo}": 2, "c{foo}": 3}) == 2
        )
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a3"), 20),
            (b("a1"), 23),
        ]

    @pytest.mark.asyncio
    async def test_zrange(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zrange("a", 0, 1) == [b("a1"), b("a2")]
        assert await r.zrange("a", 1, 2) == [b("a2"), b("a3")]

        # withscores
        assert await r.zrange("a", 0, 1, withscores=True) == [
            (b("a1"), 1.0),
            (b("a2"), 2.0),
        ]
        assert await r.zrange("a", 1, 2, withscores=True) == [
            (b("a2"), 2.0),
            (b("a3"), 3.0),
        ]

        # custom score function
        assert await r.zrange("a", 0, 1, withscores=True, score_cast_func=int) == [
            (b("a1"), 1),
            (b("a2"), 2),
        ]

    @pytest.mark.asyncio
    async def test_zrangebylex(self, r):
        await r.flushdb()
        await r.zadd("a", a=0, b=0, c=0, d=0, e=0, f=0, g=0)
        assert await r.zrangebylex("a", "-", "[c") == [b("a"), b("b"), b("c")]
        assert await r.zrangebylex("a", "-", "(c") == [b("a"), b("b")]
        assert await r.zrangebylex("a", "[aaa", "(g") == [
            b("b"),
            b("c"),
            b("d"),
            b("e"),
            b("f"),
        ]
        assert await r.zrangebylex("a", "[f", "+") == [b("f"), b("g")]
        assert await r.zrangebylex("a", "-", "+", start=3, num=2) == [b("d"), b("e")]

    @pytest.mark.asyncio
    async def test_zrangebyscore(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3, a4=4, a5=5)
        assert await r.zrangebyscore("a", 2, 4) == [b("a2"), b("a3"), b("a4")]

        # slicing with start/num
        assert await r.zrangebyscore("a", 2, 4, start=1, num=2) == [b("a3"), b("a4")]

        # withscores
        assert await r.zrangebyscore("a", 2, 4, withscores=True) == [
            (b("a2"), 2.0),
            (b("a3"), 3.0),
            (b("a4"), 4.0),
        ]

        # custom score function
        assert await r.zrangebyscore(
            "a", 2, 4, withscores=True, score_cast_func=int
        ) == [(b("a2"), 2), (b("a3"), 3), (b("a4"), 4)]

    @pytest.mark.asyncio
    async def test_zrank(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3, a4=4, a5=5)
        assert await r.zrank("a", "a1") == 0
        assert await r.zrank("a", "a2") == 1
        assert await r.zrank("a", "a6") is None

    @pytest.mark.asyncio
    async def test_zrem(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zrem("a", "a2") == 1
        assert await r.zrange("a", 0, -1) == [b("a1"), b("a3")]
        assert await r.zrem("a", "b") == 0
        assert await r.zrange("a", 0, -1) == [b("a1"), b("a3")]

    @pytest.mark.asyncio
    async def test_zrem_multiple_keys(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zrem("a", "a1", "a2") == 2
        assert await r.zrange("a", 0, 5) == [b("a3")]

    @pytest.mark.asyncio
    async def test_zremrangebylex(self, r):
        await r.flushdb()
        await r.zadd("a", a=0, b=0, c=0, d=0, e=0, f=0, g=0)
        assert await r.zremrangebylex("a", "-", "[c") == 3
        assert await r.zrange("a", 0, -1) == [b("d"), b("e"), b("f"), b("g")]
        assert await r.zremrangebylex("a", "[f", "+") == 2
        assert await r.zrange("a", 0, -1) == [b("d"), b("e")]
        assert await r.zremrangebylex("a", "[h", "+") == 0
        assert await r.zrange("a", 0, -1) == [b("d"), b("e")]

    @pytest.mark.asyncio
    async def test_zremrangebyrank(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3, a4=4, a5=5)
        assert await r.zremrangebyrank("a", 1, 3) == 3
        assert await r.zrange("a", 0, 5) == [b("a1"), b("a5")]

    @pytest.mark.asyncio
    async def test_zremrangebyscore(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3, a4=4, a5=5)
        assert await r.zremrangebyscore("a", 2, 4) == 3
        assert await r.zrange("a", 0, -1) == [b("a1"), b("a5")]
        assert await r.zremrangebyscore("a", 2, 4) == 0
        assert await r.zrange("a", 0, -1) == [b("a1"), b("a5")]

    @pytest.mark.asyncio
    async def test_zrevrange(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zrevrange("a", 0, 1) == [b("a3"), b("a2")]
        assert await r.zrevrange("a", 1, 2) == [b("a2"), b("a1")]

        # withscores
        assert await r.zrevrange("a", 0, 1, withscores=True) == [
            (b("a3"), 3.0),
            (b("a2"), 2.0),
        ]
        assert await r.zrevrange("a", 1, 2, withscores=True) == [
            (b("a2"), 2.0),
            (b("a1"), 1.0),
        ]

        # custom score function
        assert await r.zrevrange("a", 0, 1, withscores=True, score_cast_func=int) == [
            (b("a3"), 3.0),
            (b("a2"), 2.0),
        ]

    @pytest.mark.asyncio
    async def test_zrevrangebyscore(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3, a4=4, a5=5)
        assert await r.zrevrangebyscore("a", 4, 2) == [b("a4"), b("a3"), b("a2")]

        # slicing with start/num
        assert await r.zrevrangebyscore("a", 4, 2, start=1, num=2) == [b("a3"), b("a2")]

        # withscores
        assert await r.zrevrangebyscore("a", 4, 2, withscores=True) == [
            (b("a4"), 4.0),
            (b("a3"), 3.0),
            (b("a2"), 2.0),
        ]

        # custom score function
        assert await r.zrevrangebyscore(
            "a", 4, 2, withscores=True, score_cast_func=int
        ) == [(b("a4"), 4), (b("a3"), 3), (b("a2"), 2)]

    @pytest.mark.asyncio
    async def test_zrevrank(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3, a4=4, a5=5)
        assert await r.zrevrank("a", "a1") == 4
        assert await r.zrevrank("a", "a2") == 3
        assert await r.zrevrank("a", "a6") is None

    @pytest.mark.asyncio
    async def test_zscore(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=2, a3=3)
        assert await r.zscore("a", "a1") == 1.0
        assert await r.zscore("a", "a2") == 2.0
        assert await r.zscore("a", "a4") is None

    @pytest.mark.asyncio
    async def test_zunionstore_fail_crossslot(self, r):
        await r.flushdb()
        await r.zadd("a", a1=1, a2=1, a3=1)
        await r.zadd("b", a1=2, a2=2, a3=2)
        await r.zadd("c", a1=6, a3=5, a4=4)
        with pytest.raises(ResponseError) as excinfo:
            await r.zunionstore("d", ["a", "b", "c"])
        assert re.search("ClusterCrossSlotError", str(excinfo))

    @pytest.mark.asyncio
    async def test_zunionstore_sum(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=1, a3=1)
        await r.zadd("b{foo}", a1=2, a2=2, a3=2)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert await r.zunionstore("d{foo}", ["a{foo}", "b{foo}", "c{foo}"]) == 4
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a2"), 3),
            (b("a4"), 4),
            (b("a3"), 8),
            (b("a1"), 9),
        ]

    @pytest.mark.asyncio
    async def test_zunionstore_max(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=1, a3=1)
        await r.zadd("b{foo}", a1=2, a2=2, a3=2)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert (
            await r.zunionstore(
                "d{foo}", ["a{foo}", "b{foo}", "c{foo}"], aggregate="MAX"
            )
            == 4
        )
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a2"), 2),
            (b("a4"), 4),
            (b("a3"), 5),
            (b("a1"), 6),
        ]

    @pytest.mark.asyncio
    async def test_zunionstore_min(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=2, a3=3)
        await r.zadd("b{foo}", a1=2, a2=2, a3=4)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert (
            await r.zunionstore(
                "d{foo}", ["a{foo}", "b{foo}", "c{foo}"], aggregate="MIN"
            )
            == 4
        )
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a1"), 1),
            (b("a2"), 2),
            (b("a3"), 3),
            (b("a4"), 4),
        ]

    @pytest.mark.asyncio
    async def test_zunionstore_with_weight(self, r):
        await r.flushdb()
        await r.zadd("a{foo}", a1=1, a2=1, a3=1)
        await r.zadd("b{foo}", a1=2, a2=2, a3=2)
        await r.zadd("c{foo}", a1=6, a3=5, a4=4)
        assert (
            await r.zunionstore("d{foo}", {"a{foo}": 1, "b{foo}": 2, "c{foo}": 3}) == 4
        )
        assert await r.zrange("d{foo}", 0, -1, withscores=True) == [
            (b("a2"), 5),
            (b("a4"), 12),
            (b("a3"), 20),
            (b("a1"), 23),
        ]

    # HYPERLOGLOG TESTS
    @pytest.mark.asyncio
    async def test_pfadd(self, r):
        await r.flushdb()
        members = set([b("1"), b("2"), b("3")])
        assert await r.pfadd("a", *members) == 1
        assert await r.pfadd("a", *members) == 0
        assert await r.pfcount("a") == len(members)

    @pytest.mark.asyncio
    @pytest.mark.xfail(reason="New pfcount in 2.10.5 currently breaks in cluster")
    async def test_pfcount(self, r):
        await r.flushdb()
        members = set([b("1"), b("2"), b("3")])
        await r.pfadd("a", *members)
        assert await r.pfcount("a") == len(members)
        members_b = set([b("2"), b("3"), b("4")])
        await r.pfadd("b", *members_b)
        assert await r.pfcount("b") == len(members_b)
        assert await r.pfcount("a", "b") == len(members_b.union(members))

    @pytest.mark.asyncio
    async def test_pfmerge(self, r):
        await r.flushdb()
        mema = set([b("1"), b("2"), b("3")])
        memb = set([b("2"), b("3"), b("4")])
        memc = set([b("5"), b("6"), b("7")])
        await r.pfadd("a", *mema)
        await r.pfadd("b", *memb)
        await r.pfadd("c", *memc)
        await r.pfmerge("d", "c", "a")
        assert await r.pfcount("d") == 6
        await r.pfmerge("d", "b")
        assert await r.pfcount("d") == 7

    # HASH COMMANDS
    @pytest.mark.asyncio
    async def test_hget_and_hset(self, r):
        await r.flushdb()
        await r.hmset("a", {"1": 1, "2": 2, "3": 3})
        assert await r.hget("a", "1") == b("1")
        assert await r.hget("a", "2") == b("2")
        assert await r.hget("a", "3") == b("3")

        # field was updated, redis returns 0
        assert await r.hset("a", "2", 5) == 0
        assert await r.hget("a", "2") == b("5")

        # field is new, redis returns 1
        assert await r.hset("a", "4", 4) == 1
        assert await r.hget("a", "4") == b("4")

        # key inside of hash that doesn't exist returns null value
        assert await r.hget("a", "b") is None

    @pytest.mark.asyncio
    async def test_hdel(self, r):
        await r.flushdb()
        await r.hmset("a", {"1": 1, "2": 2, "3": 3})
        assert await r.hdel("a", "2") == 1
        assert await r.hget("a", "2") is None
        assert await r.hdel("a", "1", "3") == 2
        assert await r.hlen("a") == 0

    @pytest.mark.asyncio
    async def test_hexists(self, r):
        await r.flushdb()
        await r.hmset("a", {"1": 1, "2": 2, "3": 3})
        assert await r.hexists("a", "1")
        assert not await r.hexists("a", "4")

    @pytest.mark.asyncio
    async def test_hgetall(self, r):
        await r.flushdb()
        h = {b("a1"): b("1"), b("a2"): b("2"), b("a3"): b("3")}
        await r.hmset("a", h)
        assert await r.hgetall("a") == h

    @pytest.mark.asyncio
    async def test_hincrby(self, r):
        await r.flushdb()
        assert await r.hincrby("a", "1") == 1
        assert await r.hincrby("a", "1", amount=2) == 3
        assert await r.hincrby("a", "1", amount=-2) == 1

    @pytest.mark.asyncio
    async def test_hincrbyfloat(self, r):
        await r.flushdb()
        assert await r.hincrbyfloat("a", "1") == 1.0
        assert await r.hincrbyfloat("a", "1") == 2.0
        assert await r.hincrbyfloat("a", "1", 1.2) == 3.2

    @pytest.mark.asyncio
    async def test_hkeys(self, r):
        await r.flushdb()
        h = {b("a1"): b("1"), b("a2"): b("2"), b("a3"): b("3")}
        await r.hmset("a", h)
        local_keys = list(iterkeys(h))
        remote_keys = await r.hkeys("a")
        assert sorted(local_keys) == sorted(remote_keys)

    @pytest.mark.asyncio
    async def test_hlen(self, r):
        await r.flushdb()
        await r.hmset("a", {"1": 1, "2": 2, "3": 3})
        assert await r.hlen("a") == 3

    @pytest.mark.asyncio
    async def test_hmget(self, r):
        await r.flushdb()
        assert await r.hmset("a", {"a": 1, "b": 2, "c": 3})
        assert await r.hmget("a", "a", "b", "c") == [b("1"), b("2"), b("3")]

    @pytest.mark.asyncio
    async def test_hmset(self, r):
        await r.flushdb()
        h = {b("a"): b("1"), b("b"): b("2"), b("c"): b("3")}
        assert await r.hmset("a", h)
        assert await r.hgetall("a") == h

    @pytest.mark.asyncio
    async def test_hsetnx(self, r):
        await r.flushdb()
        # Initially set the hash field
        assert await r.hsetnx("a", "1", 1)
        assert await r.hget("a", "1") == b("1")
        assert not await r.hsetnx("a", "1", 2)
        assert await r.hget("a", "1") == b("1")

    @pytest.mark.asyncio
    async def test_hvals(self, r):
        await r.flushdb()
        h = {b("a1"): b("1"), b("a2"): b("2"), b("a3"): b("3")}
        await r.hmset("a", h)
        local_vals = list(itervalues(h))
        remote_vals = await r.hvals("a")
        assert sorted(local_vals) == sorted(remote_vals)

    # SORT
    @pytest.mark.asyncio
    async def test_sort_basic(self, r):
        await r.flushdb()
        await r.rpush("a", "3", "2", "1", "4")
        assert await r.sort("a") == [b("1"), b("2"), b("3"), b("4")]

    @pytest.mark.asyncio
    async def test_sort_limited(self, r):
        await r.flushdb()
        await r.rpush("a", "3", "2", "1", "4")
        assert await r.sort("a", start=1, num=2) == [b("2"), b("3")]

    @pytest.mark.asyncio
    async def test_sort_by(self, r):
        await r.flushdb()
        await r.set("score:1", 8)
        await r.set("score:2", 3)
        await r.set("score:3", 5)
        await r.rpush("a", "3", "2", "1")
        assert await r.sort("a", by="score:*") == [b("2"), b("3"), b("1")]

    @pytest.mark.asyncio
    async def test_sort_get(self, r):
        await r.flushdb()
        await r.set("user:1", "u1")
        await r.set("user:2", "u2")
        await r.set("user:3", "u3")
        await r.rpush("a", "2", "3", "1")
        assert await r.sort("a", get="user:*") == [b("u1"), b("u2"), b("u3")]

    @pytest.mark.asyncio
    async def test_sort_get_multi(self, r):
        await r.flushdb()
        await r.set("user:1", "u1")
        await r.set("user:2", "u2")
        await r.set("user:3", "u3")
        await r.rpush("a", "2", "3", "1")
        assert await r.sort("a", get=("user:*", "#")) == [
            b("u1"),
            b("1"),
            b("u2"),
            b("2"),
            b("u3"),
            b("3"),
        ]

    @pytest.mark.asyncio
    async def test_sort_get_groups_two(self, r):
        await r.flushdb()
        await r.set("user:1", "u1")
        await r.set("user:2", "u2")
        await r.set("user:3", "u3")
        await r.rpush("a", "2", "3", "1")
        assert await r.sort("a", get=("user:*", "#"), groups=True) == [
            (b("u1"), b("1")),
            (b("u2"), b("2")),
            (b("u3"), b("3")),
        ]

    @pytest.mark.asyncio
    async def test_sort_groups_string_get(self, r):
        await r.flushdb()
        await r.set("user:1", "u1")
        await r.set("user:2", "u2")
        await r.set("user:3", "u3")
        await r.rpush("a", "2", "3", "1")
        with pytest.raises(DataError):
            await r.sort("a", get="user:*", groups=True)

    @pytest.mark.asyncio
    async def test_sort_groups_just_one_get(self, r):
        await r.flushdb()
        await r.set("user:1", "u1")
        await r.set("user:2", "u2")
        await r.set("user:3", "u3")
        await r.rpush("a", "2", "3", "1")
        with pytest.raises(DataError):
            await r.sort("a", get=["user:*"], groups=True)

    @pytest.mark.asyncio
    async def test_sort_groups_no_get(self, r):
        await r.flushdb()
        await r.set("user:1", "u1")
        await r.set("user:2", "u2")
        await r.set("user:3", "u3")
        await r.rpush("a", "2", "3", "1")
        with pytest.raises(DataError):
            await r.sort("a", groups=True)

    @pytest.mark.asyncio
    async def test_sort_groups_three_gets(self, r):
        await r.flushdb()
        await r.set("user:1", "u1")
        await r.set("user:2", "u2")
        await r.set("user:3", "u3")
        await r.set("door:1", "d1")
        await r.set("door:2", "d2")
        await r.set("door:3", "d3")
        await r.rpush("a", "2", "3", "1")
        assert await r.sort("a", get=("user:*", "door:*", "#"), groups=True) == [
            (b("u1"), b("d1"), b("1")),
            (b("u2"), b("d2"), b("2")),
            (b("u3"), b("d3"), b("3")),
        ]

    @pytest.mark.asyncio
    async def test_sort_desc(self, r):
        await r.flushdb()
        await r.rpush("a", "2", "3", "1")
        assert await r.sort("a", desc=True) == [b("3"), b("2"), b("1")]

    @pytest.mark.asyncio
    async def test_sort_alpha(self, r):
        await r.flushdb()
        await r.rpush("a", "e", "c", "b", "d", "a")
        assert await r.sort("a", alpha=True) == [b("a"), b("b"), b("c"), b("d"), b("e")]

    @pytest.mark.asyncio
    async def test_sort_store(self, r):
        await r.flushdb()
        await r.rpush("a", "2", "3", "1")
        assert await r.sort("a", store="sorted_values") == 3
        assert await r.lrange("sorted_values", 0, -1) == [b("1"), b("2"), b("3")]

    @pytest.mark.asyncio
    async def test_sort_all_options(self, r):
        await r.flushdb()
        await r.set("user:1:username", "zeus")
        await r.set("user:2:username", "titan")
        await r.set("user:3:username", "hermes")
        await r.set("user:4:username", "hercules")
        await r.set("user:5:username", "apollo")
        await r.set("user:6:username", "athena")
        await r.set("user:7:username", "hades")
        await r.set("user:8:username", "dionysus")

        await r.set("user:1:favorite_drink", "yuengling")
        await r.set("user:2:favorite_drink", "rum")
        await r.set("user:3:favorite_drink", "vodka")
        await r.set("user:4:favorite_drink", "milk")
        await r.set("user:5:favorite_drink", "pinot noir")
        await r.set("user:6:favorite_drink", "water")
        await r.set("user:7:favorite_drink", "gin")
        await r.set("user:8:favorite_drink", "apple juice")

        await r.rpush("gods", "5", "8", "3", "1", "2", "7", "6", "4")
        num = await r.sort(
            "gods",
            start=2,
            num=4,
            by="user:*:username",
            get="user:*:favorite_drink",
            desc=True,
            alpha=True,
            store="sorted",
        )
        assert num == 4
        assert await r.lrange("sorted", 0, 10) == [
            b("vodka"),
            b("milk"),
            b("gin"),
            b("apple juice"),
        ]


class TestStrictCommands:
    @pytest.mark.asyncio
    async def test_strict_zadd(self, sr):
        await sr.flushdb()
        await sr.zadd("a", 1.0, "a1", 2.0, "a2", a3=3.0)
        assert await sr.zrange("a", 0, -1, withscores=True) == [
            (b("a1"), 1.0),
            (b("a2"), 2.0),
            (b("a3"), 3.0),
        ]

    @pytest.mark.asyncio
    async def test_strict_lrem(self, sr):
        await sr.flushdb()
        await sr.rpush("a", "a1", "a2", "a3", "a1")
        await sr.lrem("a", 0, "a1")
        assert await sr.lrange("a", 0, -1) == [b("a2"), b("a3")]

    @pytest.mark.asyncio
    async def test_strict_setex(self, sr):
        await sr.flushdb()
        assert await sr.setex("a", 60, "1")
        assert await sr.get("a") == b("1")
        assert 0 < await sr.ttl("a") <= 60

    @pytest.mark.asyncio
    async def test_strict_ttl(self, sr):
        await sr.flushdb()
        assert not await sr.expire("a", 10)
        await sr.set("a", "1")
        assert await sr.expire("a", 10)
        assert 0 < await sr.ttl("a") <= 10
        assert await sr.persist("a")
        assert await sr.ttl("a") == -1

    @pytest.mark.asyncio
    async def test_strict_pttl(self, sr):
        await sr.flushdb()
        assert not await sr.pexpire("a", 10000)
        await sr.set("a", "1")
        assert await sr.pexpire("a", 10000)
        assert 0 < await sr.pttl("a") <= 10000
        assert await sr.persist("a")
        assert await sr.pttl("a") == -1

    @pytest.mark.asyncio
    async def test_eval(self, sr):
        await sr.flushdb()
        res = await sr.eval(
            "return {KEYS[1],KEYS[2],ARGV[1],ARGV[2]}",
            2,
            "A{foo}",
            "B{foo}",
            "first",
            "second",
        )
        assert res[0] == b("A{foo}")
        assert res[1] == b("B{foo}")
        assert res[2] == b("first")
        assert res[3] == b("second")


class TestBinarySave:
    @pytest.mark.asyncio
    async def test_binary_get_set(self, r):
        await r.flushdb()
        assert await r.set(" foo bar ", "123")
        assert await r.get(" foo bar ") == b("123")

        assert await r.set(" foo\r\nbar\r\n ", "456")
        assert await r.get(" foo\r\nbar\r\n ") == b("456")

        assert await r.set(" \r\n\t\x07\x13 ", "789")
        assert await r.get(" \r\n\t\x07\x13 ") == b("789")

        assert sorted(await r.keys("*")) == [
            b(" \r\n\t\x07\x13 "),
            b(" foo\r\nbar\r\n "),
            b(" foo bar "),
        ]

        assert await r.delete(" foo bar ")
        assert await r.delete(" foo\r\nbar\r\n ")
        assert await r.delete(" \r\n\t\x07\x13 ")

    @pytest.mark.asyncio
    async def test_binary_lists(self, r):
        await r.flushdb()
        mapping = {
            b("foo bar"): [b("1"), b("2"), b("3")],
            b("foo\r\nbar\r\n"): [b("4"), b("5"), b("6")],
            b("foo\tbar\x07"): [b("7"), b("8"), b("9")],
        }
        # fill in lists
        for key, value in iteritems(mapping):
            await r.rpush(key, *value)

        # check that KEYS returns all the keys as they are
        assert sorted(await r.keys("*")) == sorted(list(iterkeys(mapping)))

        # check that it is possible to get list content by key name
        for key, value in iteritems(mapping):
            assert await r.lrange(key, 0, -1) == value

    @pytest.mark.asyncio
    async def test_22_info(self):
        """
        Older Redis versions contained 'allocation_stats' in INFO that
        was the cause of a number of bugs when parsing.
        """
        info = (
            b"allocation_stats:6=1,7=1,8=7141,9=180,10=92,11=116,12=5330,"
            b"13=123,14=3091,15=11048,16=225842,17=1784,18=814,19=12020,"
            b"20=2530,21=645,22=15113,23=8695,24=142860,25=318,26=3303,"
            b"27=20561,28=54042,29=37390,30=1884,31=18071,32=31367,33=160,"
            b"34=169,35=201,36=10155,37=1045,38=15078,39=22985,40=12523,"
            b"41=15588,42=265,43=1287,44=142,45=382,46=945,47=426,48=171,"
            b"49=56,50=516,51=43,52=41,53=46,54=54,55=75,56=647,57=332,"
            b"58=32,59=39,60=48,61=35,62=62,63=32,64=221,65=26,66=30,"
            b"67=36,68=41,69=44,70=26,71=144,72=169,73=24,74=37,75=25,"
            b"76=42,77=21,78=126,79=374,80=27,81=40,82=43,83=47,84=46,"
            b"85=114,86=34,87=37,88=7240,89=34,90=38,91=18,92=99,93=20,"
            b"94=18,95=17,96=15,97=22,98=18,99=69,100=17,101=22,102=15,"
            b"103=29,104=39,105=30,106=70,107=22,108=21,109=26,110=52,"
            b"111=45,112=33,113=67,114=41,115=44,116=48,117=53,118=54,"
            b"119=51,120=75,121=44,122=57,123=44,124=66,125=56,126=52,"
            b"127=81,128=108,129=70,130=50,131=51,132=53,133=45,134=62,"
            b"135=12,136=13,137=7,138=15,139=21,140=11,141=20,142=6,143=7,"
            b"144=11,145=6,146=16,147=19,148=1112,149=1,151=83,154=1,"
            b"155=1,156=1,157=1,160=1,161=1,162=2,166=1,169=1,170=1,171=2,"
            b"172=1,174=1,176=2,177=9,178=34,179=73,180=30,181=1,185=3,"
            b"187=1,188=1,189=1,192=1,196=1,198=1,200=1,201=1,204=1,205=1,"
            b"207=1,208=1,209=1,214=2,215=31,216=78,217=28,218=5,219=2,"
            b"220=1,222=1,225=1,227=1,234=1,242=1,250=1,252=1,253=1,"
            b">=256=203"
        )
        parsed = parse_info(info)
        assert "allocation_stats" in parsed
        assert "6" in parsed["allocation_stats"]
        assert ">=256" in parsed["allocation_stats"]

    @pytest.mark.asyncio
    async def test_large_responses(self, r):
        await r.flushdb()
        "The PythonParser has some special cases for return values > 1MB"
        # load up 100K of data into a key
        data = "".join([ascii_letters] * (100000 // len(ascii_letters)))
        await r.set("a", data)
        assert await r.get("a") == b(data)

    @pytest.mark.asyncio
    async def test_floating_point_encoding(self, r):
        """
        High precision floating point values sent to the server should keep
        precision.
        """
        await r.flushdb()
        timestamp = 1349673917.939762
        await r.zadd("a", timestamp, "a1")
        assert await r.zscore("a", "a1") == timestamp



from coredis.exceptions import DataError
from coredis.utils import (
    b,
    dict_merge,
    iteritems,
    first_key,
    string_keys_to_dict,
    list_or_args,
    pairs_to_dict,
)


def parse_hscan(response, **options):
    cursor, r = response
    return int(cursor), r and pairs_to_dict(r) or {}


class HashCommandMixin:

    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("HDEL HLEN", int),
        string_keys_to_dict("HEXISTS HMSET", bool),
        {
            "HGETALL": lambda r: r and pairs_to_dict(r) or {},
            "HINCRBYFLOAT": float,
            "HSCAN": parse_hscan,
        },
    )

    async def hdel(self, name, *keys):
        """Deletes ``keys`` from hash ``name``"""
        return await self.execute_command("HDEL", name, *keys)

    async def hexists(self, name, key):
        """
        Returns a boolean indicating if ``key`` exists within hash ``name``
        """
        return await self.execute_command("HEXISTS", name, key)

    async def hget(self, name, key):
        """Returns the value of ``key`` within the hash ``name``"""
        return await self.execute_command("HGET", name, key)

    async def hgetall(self, name):
        """Returns a Python dict of the hash's name/value pairs"""
        return await self.execute_command("HGETALL", name)

    async def hincrby(self, name, key, amount=1):
        """Increments the value of ``key`` in hash ``name`` by ``amount``"""
        return await self.execute_command("HINCRBY", name, key, amount)

    async def hincrbyfloat(self, name, key, amount=1.0):
        """
        Increments the value of ``key`` in hash ``name`` by floating
        ``amount``
        """
        return await self.execute_command("HINCRBYFLOAT", name, key, amount)

    async def hkeys(self, name):
        """Returns the list of keys within hash ``name``"""
        return await self.execute_command("HKEYS", name)

    async def hlen(self, name):
        """Returns the number of elements in hash ``name``"""
        return await self.execute_command("HLEN", name)

    async def hset(self, name, key, value):
        """
        Sets ``key`` to ``value`` within hash ``name``
        Returns 1 if HSET created a new field, otherwise 0
        """
        return await self.execute_command("HSET", name, key, value)

    async def hsetnx(self, name, key, value):
        """
        Sets ``key`` to ``value`` within hash ``name`` if ``key`` does not
        exist.  Returns 1 if HSETNX created a field, otherwise 0.
        """
        return await self.execute_command("HSETNX", name, key, value)

    async def hmset(self, name, mapping):
        """
        Sets key to value within hash ``name`` for each corresponding
        key and value from the ``mapping`` dict.
        """
        if not mapping:
            raise DataError("'hmset' with 'mapping' of length 0")
        items = []
        for pair in iteritems(mapping):
            items.extend(pair)
        return await self.execute_command("HMSET", name, *items)

    async def hmget(self, name, keys, *args):
        """Returns a list of values ordered identically to ``keys``"""
        args = list_or_args(keys, args)
        return await self.execute_command("HMGET", name, *args)

    async def hvals(self, name):
        """Returns the list of values within hash ``name``"""
        return await self.execute_command("HVALS", name)

    async def hscan(self, name, cursor=0, match=None, count=None):
        """
        Incrementallys return key/value slices in a hash. Also returns a
        cursor pointing to the scan position.

        ``match`` allows for filtering the keys by pattern

        ``count`` allows for hint the minimum number of returns
        """
        pieces = [name, cursor]
        if match is not None:
            pieces.extend([b("MATCH"), match])
        if count is not None:
            pieces.extend([b("COUNT"), count])
        return await self.execute_command("HSCAN", *pieces)

    async def hstrlen(self, name, key):
        """
        Returns the string length of the value associated
        with field in the hash stored at key.
        If the key or the field do not exist, 0 is returned.
        """
        return await self.execute_command("HSTRLEN", name, key)


class ClusterHashCommandMixin(HashCommandMixin):

    RESULT_CALLBACKS = {"HSCAN": first_key}



#!/usr/bin/python
# -*- coding: utf-8 -*-
import coredis
import asyncio
import pytest
import sys
from unittest.mock import Mock

_REDIS_VERSIONS = {}


async def get_version(**kwargs):
    params = {"host": "localhost", "port": 6379, "db": 0}
    params.update(kwargs)
    key = "%s:%s" % (params["host"], params["port"])
    if key not in _REDIS_VERSIONS:
        client = coredis.StrictRedis(**params)
        _REDIS_VERSIONS[key] = (await client.info())["redis_version"]
        client.connection_pool.disconnect()
    return _REDIS_VERSIONS[key]


@pytest.fixture()
def r(event_loop):
    return coredis.StrictRedis(loop=event_loop)


class AsyncMock(Mock):
    def __init__(self, *args, **kwargs):
        super(AsyncMock, self).__init__(*args, **kwargs)

    def __await__(self):
        future = asyncio.Future(loop=self.loop)
        future.set_result(self)
        result = yield from future
        return result

    @staticmethod
    def pack_response(response, *, loop):
        future = asyncio.Future(loop=loop)
        future.set_result(response)
        return future


def _gen_mock_resp(r, response, *, loop):
    mock_connection_pool = AsyncMock(loop=loop)
    connection = AsyncMock(loop=loop)
    connection.read_response.return_value = AsyncMock.pack_response(response, loop=loop)
    mock_connection_pool.get_connection.return_value = connection
    r.connection_pool = mock_connection_pool
    return r


@pytest.fixture()
def mock_resp_role(event_loop):
    r = coredis.StrictRedis(loop=event_loop)
    response = [b"master", 169, [[b"172.17.0.2", b"7004", b"169"]]]
    return _gen_mock_resp(r, response, loop=event_loop)



from coredis.client import StrictRedis, StrictRedisCluster
from coredis.connection import Connection, UnixDomainSocketConnection, ClusterConnection
from coredis.pool import ConnectionPool, ClusterConnectionPool
from coredis.exceptions import (
    AuthenticationError,
    BusyLoadingError,
    ConnectionError,
    DataError,
    InvalidResponse,
    PubSubError,
    ReadOnlyError,
    RedisError,
    ResponseError,
    TimeoutError,
    WatchError,
    CompressError,
    ClusterDownException,
    ClusterCrossSlotError,
    CacheError,
    ClusterDownError,
    ClusterError,
    RedisClusterException,
    RedisClusterError,
    ExecAbortError,
    LockError,
    NoScriptError,
)


__version__ = "1.1.8"

VERSION = tuple(map(int, __version__.split(".")))


__all__ = [
    "StrictRedis",
    "StrictRedisCluster",
    "Connection",
    "UnixDomainSocketConnection",
    "ClusterConnection",
    "ConnectionPool",
    "ClusterConnectionPool",
    "AuthenticationError",
    "BusyLoadingError",
    "ConnectionError",
    "DataError",
    "InvalidResponse",
    "PubSubError",
    "ReadOnlyError",
    "RedisError",
    "ResponseError",
    "TimeoutError",
    "WatchError",
    "CompressError",
    "ClusterDownException",
    "ClusterCrossSlotError",
    "CacheError",
    "ClusterDownError",
    "ClusterError",
    "RedisClusterException",
    "RedisClusterError",
    "ExecAbortError",
    "LockError",
    "NoScriptError",
]

from . import _version

__version__ = _version.get_versions()["version"]



import asyncio
import sys

from coredis.commands.cluster import ClusterCommandMixin
from coredis.commands.connection import (
    ClusterConnectionCommandMixin,
    ConnectionCommandMixin,
)
from coredis.commands.extra import ExtraCommandMixin
from coredis.commands.geo import GeoCommandMixin
from coredis.commands.hash import ClusterHashCommandMixin, HashCommandMixin
from coredis.commands.hyperlog import ClusterHyperLogCommandMixin, HyperLogCommandMixin
from coredis.commands.keys import ClusterKeysCommandMixin, KeysCommandMixin
from coredis.commands.lists import ClusterListsCommandMixin, ListsCommandMixin
from coredis.commands.pubsub import CLusterPubSubCommandMixin, PubSubCommandMixin
from coredis.commands.scripting import (
    ClusterScriptingCommandMixin,
    ScriptingCommandMixin,
)
from coredis.commands.sentinel import ClusterSentinelCommands, SentinelCommandMixin
from coredis.commands.server import ClusterServerCommandMixin, ServerCommandMixin
from coredis.commands.sets import ClusterSetsCommandMixin, SetsCommandMixin
from coredis.commands.sorted_set import (
    ClusterSortedSetCommandMixin,
    SortedSetCommandMixin,
)
from coredis.commands.streams import StreamsCommandMixin
from coredis.commands.strings import ClusterStringsCommandMixin, StringsCommandMixin
from coredis.commands.transaction import (
    ClusterTransactionCommandMixin,
    TransactionCommandMixin,
)
from coredis.compat import CancelledError
from coredis.connection import RedisSSLContext, UnixDomainSocketConnection
from coredis.exceptions import (
    AskError,
    BusyLoadingError,
    ClusterDownError,
    ClusterError,
    ConnectionError,
    MovedError,
    RedisClusterException,
    TimeoutError,
    TryAgainError,
)
from coredis.pool import ClusterConnectionPool, ConnectionPool
from coredis.utils import (
    NodeFlag,
    blocked_command,
    clusterdown_wrapper,
    dict_merge,
    first_key,
)

mixins = [
    ClusterCommandMixin,
    ConnectionCommandMixin,
    ExtraCommandMixin,
    GeoCommandMixin,
    HashCommandMixin,
    HyperLogCommandMixin,
    KeysCommandMixin,
    ListsCommandMixin,
    PubSubCommandMixin,
    ScriptingCommandMixin,
    SentinelCommandMixin,
    ServerCommandMixin,
    SetsCommandMixin,
    SortedSetCommandMixin,
    StringsCommandMixin,
    TransactionCommandMixin,
    StreamsCommandMixin,
]

cluster_mixins = [
    ClusterCommandMixin,
    ClusterStringsCommandMixin,
    ClusterServerCommandMixin,
    ClusterConnectionCommandMixin,
    CLusterPubSubCommandMixin,
    ClusterSentinelCommands,
    ClusterKeysCommandMixin,
    ClusterScriptingCommandMixin,
    ClusterHashCommandMixin,
    ClusterSetsCommandMixin,
    ClusterSortedSetCommandMixin,
    ClusterTransactionCommandMixin,
    ClusterListsCommandMixin,
    ClusterHyperLogCommandMixin,
]

if sys.version_info[:2] >= (3, 6):
    from coredis.commands.iter import IterCommandMixin, ClusterIterCommandMixin

    mixins.append(IterCommandMixin)
    cluster_mixins.append(ClusterIterCommandMixin)


class StrictRedis(*mixins):
    """
    Implementation of the Redis protocol.

    This abstract class provides a Python interface to all Redis commands
    and an implementation of the Redis protocol.

    Connection and Pipeline derive from this, implementing how
    the commands are sent and received to the Redis server
    """

    RESPONSE_CALLBACKS = dict_merge(*[mixin.RESPONSE_CALLBACKS for mixin in mixins])

    @classmethod
    def from_url(cls, url, db=None, **kwargs):
        """
        Return a Redis client object configured from the given URL, which must
        use either `the ``redis://`` scheme
        <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ for RESP
        connections or the ``unix://`` scheme for Unix domain sockets.

        For example:

        redis://[:password]@localhost:6379/0
        unix://[:password]@/path/to/socket.sock?db=0

        There are several ways to specify a database number. The parse function
        will return the first specified option:
        1. A ``db`` querystring option, e.g. redis://localhost?db=0
        2. If using the redis:// scheme, the path argument of the url, e.g.
        redis://localhost/0
        3. The ``db`` argument to this function.

        If none of these options are specified, db=0 is used.

        Any additional querystring arguments and keyword arguments will be
        passed along to the ConnectionPool class's initializer. In the case
        of conflicting arguments, querystring arguments always win.
        """
        connection_pool = ConnectionPool.from_url(url, db=db, **kwargs)
        return cls(connection_pool=connection_pool)

    def __init__(
        self,
        host="localhost",
        port=6379,
        db=0,
        password=None,
        stream_timeout=None,
        connect_timeout=None,
        connection_pool=None,
        unix_socket_path=None,
        encoding="utf-8",
        decode_responses=False,
        ssl=False,
        ssl_context=None,
        ssl_keyfile=None,
        ssl_certfile=None,
        ssl_cert_reqs=None,
        ssl_ca_certs=None,
        max_connections=None,
        retry_on_timeout=False,
        max_idle_time=0,
        idle_check_interval=1,
        loop=None,
        **kwargs
    ):
        if not connection_pool:
            kwargs = {
                "db": db,
                "password": password,
                "encoding": encoding,
                "stream_timeout": stream_timeout,
                "connect_timeout": connect_timeout,
                "max_connections": max_connections,
                "retry_on_timeout": retry_on_timeout,
                "decode_responses": decode_responses,
                "max_idle_time": max_idle_time,
                "idle_check_interval": idle_check_interval,
                "loop": loop,
            }
            # based on input, setup appropriate connection args
            if unix_socket_path is not None:
                kwargs.update(
                    {
                        "path": unix_socket_path,
                        "connection_class": UnixDomainSocketConnection,
                    }
                )
            else:
                # TCP specific options
                kwargs.update({"host": host, "port": port})
                if ssl_context is not None:
                    kwargs["ssl_context"] = ssl_context
                elif ssl:
                    ssl_context = RedisSSLContext(
                        ssl_keyfile, ssl_certfile, ssl_cert_reqs, ssl_ca_certs
                    ).get()
                    kwargs["ssl_context"] = ssl_context
            connection_pool = ConnectionPool(**kwargs)
        self.connection_pool = connection_pool
        self._use_lua_lock = None

        self.response_callbacks = self.__class__.RESPONSE_CALLBACKS.copy()

    def __repr__(self):
        return "{}<{}>".format(type(self).__name__, repr(self.connection_pool))

    def set_response_callback(self, command, callback):
        """Sets a custom Response Callback"""
        self.response_callbacks[command] = callback

    # COMMAND EXECUTION AND PROTOCOL PARSING
    async def execute_command(self, *args, **options):
        """Executes a command and returns a parsed response"""
        pool = self.connection_pool
        command_name = args[0]
        connection = pool.get_connection()
        try:
            await connection.send_command(*args)
            return await self.parse_response(connection, command_name, **options)
        except CancelledError:
            # do not retry when coroutine is cancelled
            connection.disconnect()
            raise
        except (ConnectionError, TimeoutError) as e:
            connection.disconnect()
            if not connection.retry_on_timeout and isinstance(e, TimeoutError):
                raise
            await connection.send_command(*args)
            return await self.parse_response(connection, command_name, **options)
        finally:
            pool.release(connection)

    async def parse_response(self, connection, command_name, **options):
        """Parses a response from the Redis server"""
        response = await connection.read_response()
        if command_name in self.response_callbacks:
            callback = self.response_callbacks[command_name]
            return callback(response, **options)
        return response

    async def pipeline(self, transaction=True, shard_hint=None):
        """
        Returns a new pipeline object that can queue multiple commands for
        later execution. ``transaction`` indicates whether all commands
        should be executed atomically. Apart from making a group of operations
        atomic, pipelines are useful for reducing the back-and-forth overhead
        between the client and server.
        """
        from coredis.pipeline import StrictPipeline

        pipeline = StrictPipeline(
            self.connection_pool, self.response_callbacks, transaction, shard_hint
        )
        await pipeline.reset()
        return pipeline


class StrictRedisCluster(StrictRedis, *cluster_mixins):
    """
    If a command is implemented over the one in StrictRedis then it requires some changes compared to
    the regular implementation of the method.
    """

    RedisClusterRequestTTL = 16
    NODES_FLAGS = dict_merge(
        *[
            mixin.NODES_FLAGS
            for mixin in cluster_mixins
            if hasattr(mixin, "NODES_FLAGS")
        ]
    )
    RESULT_CALLBACKS = dict_merge(
        *[
            mixin.RESULT_CALLBACKS
            for mixin in cluster_mixins
            if hasattr(mixin, "RESULT_CALLBACKS")
        ]
    )

    def __init__(
        self,
        host=None,
        port=None,
        startup_nodes=None,
        max_connections=32,
        max_connections_per_node=False,
        readonly=False,
        reinitialize_steps=None,
        skip_full_coverage_check=False,
        nodemanager_follow_cluster=False,
        **kwargs
    ):
        """
        :startup_nodes:
        List of nodes that initial bootstrapping can be done from
        :host:
        Can be used to point to a startup node
        :port:
        Can be used to point to a startup node
        :max_connections:
        Maximum number of connections that should be kept open at one time
        :readonly:
        enable READONLY mode. You can read possibly stale data from slave.
        :skip_full_coverage_check:
        Skips the check of cluster-require-full-coverage config, useful for clusters
        without the CONFIG command (like aws)
        :nodemanager_follow_cluster:
        The node manager will during initialization try the last set of nodes that
        it was operating on. This will allow the client to drift along side the cluster
        if the cluster nodes move around alot.
        :**kwargs:
        Extra arguments that will be sent into StrictRedis instance when created
        (See Official redis-py doc for supported kwargs
        [https://github.com/andymccurdy/redis-py/blob/master/redis/client.py])
        Some kwargs is not supported and will raise RedisClusterException
        - db (Redis do not support database SELECT in cluster mode)
        """
        # Tweaks to StrictRedis client arguments when running in cluster mode
        if "db" in kwargs:
            raise RedisClusterException(
                "Argument 'db' is not possible to use in cluster mode"
            )
        if "connection_pool" in kwargs:
            pool = kwargs.pop("connection_pool")
        else:
            startup_nodes = [] if startup_nodes is None else startup_nodes

            # Support host/port as argument
            if host:
                startup_nodes.append({"host": host, "port": port if port else 7000})
            pool = ClusterConnectionPool(
                startup_nodes=startup_nodes,
                max_connections=max_connections,
                reinitialize_steps=reinitialize_steps,
                max_connections_per_node=max_connections_per_node,
                skip_full_coverage_check=skip_full_coverage_check,
                nodemanager_follow_cluster=nodemanager_follow_cluster,
                readonly=readonly,
                **kwargs
            )

        super(StrictRedisCluster, self).__init__(connection_pool=pool, **kwargs)

        self.refresh_table_asap = False
        self.nodes_flags = self.__class__.NODES_FLAGS.copy()
        self.result_callbacks = self.__class__.RESULT_CALLBACKS.copy()
        self.response_callbacks = self.__class__.RESPONSE_CALLBACKS.copy()

    @classmethod
    def from_url(cls, url, db=None, skip_full_coverage_check=False, **kwargs):
        """
        Return a Redis client object configured from the given URL, which must
        use either `the ``redis://`` scheme
        <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ for RESP
        connections or the ``unix://`` scheme for Unix domain sockets.
        For example::
        redis://[:password]@localhost:6379/0
        unix://[:password]@/path/to/socket.sock?db=0
        There are several ways to specify a database number. The parse function
        will return the first specified option:
        1. A ``db`` querystring option, e.g. redis://localhost?db=0
        2. If using the redis:// scheme, the path argument of the url, e.g.
        redis://localhost/0
        3. The ``db`` argument to this function.
        If none of these options are specified, db=0 is used.
        Any additional querystring arguments and keyword arguments will be
        passed along to the ConnectionPool class's initializer. In the case
        of conflicting arguments, querystring arguments always win.
        """
        connection_pool = ClusterConnectionPool.from_url(
            url, db=db, skip_full_coverage_check=skip_full_coverage_check, **kwargs
        )
        return cls(connection_pool=connection_pool)

    def __repr__(self):
        servers = list(
            {
                "{0}:{1}".format(info["host"], info["port"])
                for info in self.connection_pool.nodes.startup_nodes
            }
        )
        servers.sort()
        return "{0}<{1}>".format(type(self).__name__, ", ".join(servers))

    def set_result_callback(self, command, callback):
        "Sets a custom Result Callback"
        self.result_callbacks[command] = callback

    def _determine_slot(self, *args):
        """Figures out what slot based on command and args"""
        if len(args) <= 1:
            raise RedisClusterException(
                "No way to dispatch this command to Redis Cluster. Missing key."
            )
        command = args[0]

        if command in ["EVAL", "EVALSHA"]:
            numkeys = args[2]
            keys = args[3 : 3 + numkeys]
            slots = {self.connection_pool.nodes.keyslot(key) for key in keys}
            if len(slots) != 1:
                raise RedisClusterException(
                    "{0} - all keys must map to the same key slot".format(command)
                )
            return slots.pop()
        elif command in ("XREAD", "XREADGROUP"):
            try:
                idx = args.index("STREAMS") + 1
            except ValueError:
                raise RedisClusterException(
                    "{0} arguments do not contain STREAMS operand".format(command)
                )
            key = args[idx]
        elif command in ("XGROUP", "XINFO"):
            key = args[2]
        else:
            key = args[1]

        return self.connection_pool.nodes.keyslot(key)

    def _merge_result(self, command, res, **kwargs):
        """
        `res` is a dict with the following structure Dict(NodeName, CommandResult)
        """
        if command in self.result_callbacks:
            return self.result_callbacks[command](res, **kwargs)

        # Default way to handle result
        return first_key(res)

    def determine_node(self, *args, **kwargs):
        """
        TODO: document
        """
        command = args[0]
        node_flag = self.nodes_flags.get(command)

        if node_flag == NodeFlag.BLOCKED:
            return blocked_command(self, command)
        elif node_flag == NodeFlag.RANDOM:
            return [self.connection_pool.nodes.random_node()]
        elif node_flag == NodeFlag.ALL_MASTERS:
            return self.connection_pool.nodes.all_masters()
        elif node_flag == NodeFlag.ALL_NODES:
            return self.connection_pool.nodes.all_nodes()
        elif node_flag == NodeFlag.SLOT_ID:
            # if node flag of command is SLOT_ID
            # `slot_id` should is assumed in kwargs
            slot = kwargs.get("slot_id")
            if not slot:
                raise RedisClusterException(
                    "slot_id is needed to execute command {}".format(command)
                )
            return [self.connection_pool.nodes.node_from_slot(slot)]
        else:
            return None

    @clusterdown_wrapper
    async def execute_command(self, *args, **kwargs):
        """
        Sends a command to a node in the cluster
        """
        if not self.connection_pool.initialized:
            await self.connection_pool.initialize()
        if not args:
            raise RedisClusterException("Unable to determine command to use")

        command = args[0]

        node = self.determine_node(*args, **kwargs)
        if node:
            return await self.execute_command_on_nodes(node, *args, **kwargs)

        # If set externally we must update it before calling any commands
        if self.refresh_table_asap:
            await self.connection_pool.nodes.initialize()
            self.refresh_table_asap = False

        redirect_addr = None
        asking = False

        try_random_node = False
        slot = self._determine_slot(*args)
        ttl = int(self.RedisClusterRequestTTL)

        while ttl > 0:
            ttl -= 1

            if asking:
                node = self.connection_pool.nodes.nodes[redirect_addr]
                r = self.connection_pool.get_connection_by_node(node)
            elif try_random_node:
                r = self.connection_pool.get_random_connection()
                try_random_node = False
            else:
                if self.refresh_table_asap:
                    # MOVED
                    node = self.connection_pool.get_master_node_by_slot(slot)
                else:
                    node = self.connection_pool.get_node_by_slot(slot)
                r = self.connection_pool.get_connection_by_node(node)

            try:
                if asking:
                    await r.send_command("ASKING")
                    await self.parse_response(r, "ASKING", **kwargs)
                    asking = False

                await r.send_command(*args)
                return await self.parse_response(r, command, **kwargs)
            except (RedisClusterException, BusyLoadingError, CancelledError):
                raise
            except (ConnectionError, TimeoutError):
                try_random_node = True

                if ttl < self.RedisClusterRequestTTL / 2:
                    await asyncio.sleep(0.1)
            except ClusterDownError as e:
                self.connection_pool.disconnect()
                self.connection_pool.reset()
                self.refresh_table_asap = True

                raise e
            except MovedError as e:
                # Reinitialize on ever x number of MovedError.
                # This counter will increase faster when the same client object
                # is shared between multiple threads. To reduce the frequency you
                # can set the variable 'reinitialize_steps' in the constructor.
                self.refresh_table_asap = True
                await self.connection_pool.nodes.increment_reinitialize_counter()

                node = self.connection_pool.nodes.set_node(
                    e.host, e.port, server_type="master"
                )
                self.connection_pool.nodes.slots[e.slot_id][0] = node
            except TryAgainError as e:
                if ttl < self.RedisClusterRequestTTL / 2:
                    await asyncio.sleep(0.05)
            except AskError as e:
                redirect_addr, asking = "{0}:{1}".format(e.host, e.port), True
            finally:
                self.connection_pool.release(r)

        raise ClusterError("TTL exhausted.")

    async def execute_command_on_nodes(self, nodes, *args, **kwargs):
        command = args[0]
        res = {}

        for node in nodes:
            connection = self.connection_pool.get_connection_by_node(node)

            # copy from redis-py
            try:
                await connection.send_command(*args)
                res[node["name"]] = await self.parse_response(
                    connection, command, **kwargs
                )
            except CancelledError:
                # do not retry when coroutine is cancelled
                connection.disconnect()
                raise
            except (ConnectionError, TimeoutError) as e:
                connection.disconnect()

                if not connection.retry_on_timeout and isinstance(e, TimeoutError):
                    raise

                await connection.send_command(*args)
                res[node["name"]] = await self.parse_response(
                    connection, command, **kwargs
                )
            finally:
                self.connection_pool.release(connection)
        return self._merge_result(command, res, **kwargs)

    async def pipeline(self, transaction=None, shard_hint=None, watches=None):
        """
        Cluster impl:
            Pipelines do not work in cluster mode the same way they do in normal mode.
            Create a clone of this object so that simulating pipelines will work correctly.
            Each command will be called directly when used and when calling execute() will only return the result stack.
        cluster transaction can only be run with commands in the same node, otherwise error will be raised.
        """
        await self.connection_pool.initialize()
        if shard_hint:
            raise RedisClusterException("shard_hint is deprecated in cluster mode")

        from coredis.pipeline import StrictClusterPipeline

        return StrictClusterPipeline(
            connection_pool=self.connection_pool,
            startup_nodes=self.connection_pool.nodes.startup_nodes,
            result_callbacks=self.result_callbacks,
            response_callbacks=self.response_callbacks,
            transaction=transaction,
            watches=watches,
        )



from __future__ import with_statement
import pytest
import pickle
import coredis


class TestEncoding:
    @pytest.fixture()
    def r(self, request):
        return coredis.StrictRedis(decode_responses=True)

    @pytest.mark.asyncio()
    async def test_simple_encoding(self, r):
        await r.flushdb()
        unicode_string = chr(124) + "abcd" + chr(125)
        await r.set("unicode-string", unicode_string)
        cached_val = await r.get("unicode-string")
        assert isinstance(cached_val, str)
        assert unicode_string == cached_val

    @pytest.mark.asyncio()
    async def test_list_encoding(self, r):
        unicode_string = chr(124) + "abcd" + chr(125)
        result = [unicode_string, unicode_string, unicode_string]
        await r.rpush("a", *result)
        assert await r.lrange("a", 0, -1) == result

    @pytest.mark.asyncio()
    async def test_object_value(self, r):
        unicode_string = chr(124) + "abcd" + chr(125)
        await r.set("unicode-string", Exception(unicode_string))
        cached_val = await r.get("unicode-string")
        assert isinstance(cached_val, str)
        assert unicode_string == cached_val

    @pytest.mark.asyncio()
    async def test_pickled_object(self):
        r = coredis.StrictRedis()
        obj = Exception("args")
        pickled_obj = pickle.dumps(obj)
        await r.set("pickled-obj", pickled_obj)
        cached_obj = await r.get("pickled-obj")
        assert isinstance(cached_obj, bytes)
        assert obj.args == pickle.loads(cached_obj).args
